#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of user tables."""

from datetime import datetime
from secrets import token_hex
from typing import Any, Optional, Sequence, Tuple
from uuid import UUID

from passlib.context import CryptContext
from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship

from zenml.models import (
    APIKeyInternalResponse,
    APIKeyInternalUpdate,
    APIKeyRequest,
    APIKeyResponse,
    APIKeyResponseBody,
    APIKeyResponseMetadata,
    APIKeyRotateRequest,
    APIKeyUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg


class APIKeySchema(NamedSchema, table=True):
    """SQL Model for API keys."""

    __tablename__ = "api_key"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "service_account_id",
            name="unique_api_key_name_in_service_account",
        ),
    )

    description: str = Field(sa_column=Column(TEXT))
    key: str
    previous_key: Optional[str] = Field(default=None, nullable=True)
    retain_period: int = Field(default=0)
    active: bool = Field(default=True)
    last_login: Optional[datetime] = None
    last_rotated: Optional[datetime] = None

    service_account_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="service_account_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    service_account: "UserSchema" = Relationship(back_populates="api_keys")

    @classmethod
    def _generate_jwt_secret_key(cls) -> str:
        """Generate a random API key.

        Returns:
            A random API key.
        """
        return token_hex(32)

    @classmethod
    def _get_hashed_key(cls, key: str) -> str:
        """Hashes the input key and returns the hash value.

        Args:
            key: The key value to hash.

        Returns:
            The key hash value.
        """
        context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        return context.hash(key)

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = [
            joinedload(jl_arg(APIKeySchema.service_account), innerjoin=True),
        ]

        return options

    @classmethod
    def from_request(
        cls,
        service_account_id: UUID,
        request: APIKeyRequest,
    ) -> Tuple["APIKeySchema", str]:
        """Convert a `APIKeyRequest` to a `APIKeySchema`.

        Args:
            service_account_id: The service account id to associate the key
                with.
            request: The request model to convert.

        Returns:
            The converted schema and the un-hashed API key.
        """
        key = cls._generate_jwt_secret_key()
        hashed_key = cls._get_hashed_key(key)
        now = utc_now()
        return (
            cls(
                name=request.name,
                description=request.description or "",
                key=hashed_key,
                service_account_id=service_account_id,
                created=now,
                updated=now,
            ),
            key,
        )

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> APIKeyResponse:
        """Convert a `APIKeySchema` to an `APIKeyResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic

            **kwargs: Keyword arguments to filter models.

        Returns:
            The created APIKeyResponse.
        """
        metadata = None
        if include_metadata:
            metadata = APIKeyResponseMetadata(
                description=self.description,
                retain_period_minutes=self.retain_period,
                last_login=self.last_login,
                last_rotated=self.last_rotated,
            )

        body = APIKeyResponseBody(
            created=self.created,
            updated=self.updated,
            active=self.active,
            service_account=self.service_account.to_service_account_model(),
        )

        return APIKeyResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
        )

    def to_internal_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
    ) -> APIKeyInternalResponse:
        """Convert a `APIKeySchema` to an `APIKeyInternalResponse`.

        The internal response model includes the hashed key values.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.

        Returns:
            The created APIKeyInternalResponse.
        """
        model = self.to_model(
            include_metadata=include_metadata,
            include_resources=include_resources,
        )
        model.get_body().key = self.key

        return APIKeyInternalResponse(
            id=self.id,
            name=self.name,
            previous_key=self.previous_key,
            body=model.body,
            metadata=model.metadata,
        )

    def update(self, update: APIKeyUpdate) -> "APIKeySchema":
        """Update an `APIKeySchema` with an `APIKeyUpdate`.

        Args:
            update: The update model.

        Returns:
            The updated `APIKeySchema`.
        """
        for field, value in update.model_dump(exclude_none=True).items():
            if hasattr(self, field):
                setattr(self, field, value)

        self.updated = utc_now()
        return self

    def internal_update(self, update: APIKeyInternalUpdate) -> "APIKeySchema":
        """Update an `APIKeySchema` with an `APIKeyInternalUpdate`.

        The internal update can also update the last used timestamp.

        Args:
            update: The update model.

        Returns:
            The updated `APIKeySchema`.
        """
        self.update(update)

        if update.update_last_login:
            self.last_login = self.updated

        return self

    def rotate(
        self,
        rotate_request: APIKeyRotateRequest,
    ) -> Tuple["APIKeySchema", str]:
        """Rotate the key for an `APIKeySchema`.

        Args:
            rotate_request: The rotate request model.

        Returns:
            The updated `APIKeySchema` and the new un-hashed key.
        """
        self.updated = utc_now()
        self.previous_key = self.key
        self.retain_period = rotate_request.retain_period_minutes
        new_key = self._generate_jwt_secret_key()
        self.key = self._get_hashed_key(new_key)
        self.last_rotated = self.updated

        return self, new_key



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base and meta classes for ZenML integrations."""

from typing import Any, Dict, List, Optional, Tuple, Type, cast

from packaging.requirements import Requirement

from zenml.integrations.registry import integration_registry
from zenml.logger import get_logger
from zenml.stack.flavor import Flavor
from zenml.utils.package_utils import get_dependencies, requirement_installed

logger = get_logger(__name__)


class IntegrationMeta(type):
    """Metaclass responsible for registering different Integration subclasses."""

    def __new__(
        mcs, name: str, bases: Tuple[Type[Any], ...], dct: Dict[str, Any]
    ) -> "IntegrationMeta":
        """Hook into creation of an Integration class.

        Args:
            name: The name of the class being created.
            bases: The base classes of the class being created.
            dct: The dictionary of attributes of the class being created.

        Returns:
            The newly created class.
        """
        cls = cast(Type["Integration"], super().__new__(mcs, name, bases, dct))
        if name != "Integration":
            integration_registry.register_integration(cls.NAME, cls)
        return cls


class Integration(metaclass=IntegrationMeta):
    """Base class for integration in ZenML."""

    NAME = "base_integration"

    REQUIREMENTS: List[str] = []
    APT_PACKAGES: List[str] = []
    REQUIREMENTS_IGNORED_ON_UNINSTALL: List[str] = []

    @classmethod
    def check_installation(cls) -> bool:
        """Method to check whether the required packages are installed.

        Returns:
            True if all required packages are installed, False otherwise.
        """
        for requirement in cls.get_requirements():
            parsed_requirement = Requirement(requirement)

            if not requirement_installed(parsed_requirement):
                logger.debug(
                    "Requirement '%s' for integration '%s' is not installed "
                    "or installed with the wrong version.",
                    requirement,
                    cls.NAME,
                )
                return False

            dependencies = get_dependencies(parsed_requirement)

            for dependency in dependencies:
                if not requirement_installed(dependency):
                    logger.debug(
                        "Requirement '%s' for integration '%s' is not "
                        "installed or installed with the wrong version.",
                        dependency,
                        cls.NAME,
                    )
                    return False

        logger.debug(
            f"Integration '{cls.NAME}' is installed correctly with "
            f"requirements {cls.get_requirements()}."
        )
        return True

    @classmethod
    def get_requirements(
        cls,
        target_os: Optional[str] = None,
        python_version: Optional[str] = None,
    ) -> List[str]:
        """Method to get the requirements for the integration.

        Args:
            target_os: The target operating system to get the requirements for.
            python_version: The Python version to use for the requirements.

        Returns:
            A list of requirements.
        """
        return cls.REQUIREMENTS

    @classmethod
    def get_uninstall_requirements(
        cls, target_os: Optional[str] = None
    ) -> List[str]:
        """Method to get the uninstall requirements for the integration.

        Args:
            target_os: The target operating system to get the requirements for.

        Returns:
            A list of requirements.
        """
        ret = []
        for each in cls.get_requirements(target_os=target_os):
            is_ignored = False
            for ignored in cls.REQUIREMENTS_IGNORED_ON_UNINSTALL:
                if each.startswith(ignored):
                    is_ignored = True
                    break
            if not is_ignored:
                ret.append(each)
        return ret

    @classmethod
    def activate(cls) -> None:
        """Abstract method to activate the integration."""

    @classmethod
    def flavors(cls) -> List[Type[Flavor]]:
        """Abstract method to declare new stack component flavors.

        Returns:
            A list of new stack component flavors.
        """
        return []



"""Pipeline run replays [bf4203afb657].

Revision ID: bf4203afb657
Revises: b949c5d1bc0d
Create Date: 2026-02-02 14:17:42.255690

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "bf4203afb657"
down_revision = "b949c5d1bc0d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "original_run_id", sqlmodel.sql.sqltypes.GUID(), nullable=True
            )
        )
        batch_op.create_foreign_key(
            "fk_pipeline_run_original_run_id_pipeline_run",
            "pipeline_run",
            ["original_run_id"],
            ["id"],
            ondelete="SET NULL",
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_pipeline_run_original_run_id_pipeline_run", type_="foreignkey"
        )
        batch_op.drop_column("original_run_id")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""FastAPI implementation of the deployment app factory and adapters."""


from typing import List, Type
from zenml.deployers.server.app import BaseDeploymentAppRunner, BaseDeploymentAppRunnerFlavor

FASTAPI_APP_RUNNER_FLAVOR_NAME = "fastapi"

class FastAPIDeploymentAppRunnerFlavor(BaseDeploymentAppRunnerFlavor):
    """FastAPI deployment app runner flavor."""

    @property
    def name(self) -> str:
        """The name of the deployment app runner flavor.

        Returns:
            The name of the deployment app runner flavor.
        """
        return FASTAPI_APP_RUNNER_FLAVOR_NAME

    @property
    def implementation_class(self) -> Type[BaseDeploymentAppRunner]:
        """The class that implements the deployment app runner.

        Returns:
            The implementation class for the deployment app runner.
        """
        from zenml.deployers.server.fastapi.app import FastAPIDeploymentAppRunner
        return FastAPIDeploymentAppRunner

    @property
    def requirements(self) -> List[str]:
        """The software requirements for the deployment app runner.

        Returns:
            The software requirements for the deployment app runner.
        """
        return super().requirements + ["fastapi"]


#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI functionality to interact with deployments."""

import json
from typing import Any, List, Optional
from uuid import UUID

import click

from zenml.cli import utils as cli_utils
from zenml.cli.cli import TagGroup, cli
from zenml.cli.utils import OutputFormat, fetch_snapshot, list_options
from zenml.client import Client
from zenml.console import console
from zenml.deployers.exceptions import DeploymentInvalidParametersError
from zenml.enums import CliCategories
from zenml.logger import get_logger
from zenml.models import (
    DeploymentFilter,
)
from zenml.pipelines.pipeline_definition import Pipeline
from zenml.utils import source_utils
from zenml.utils.dashboard_utils import get_deployment_url

logger = get_logger(__name__)


def _import_pipeline(source: str) -> Pipeline:
    """Import a pipeline.

    Args:
        source: The pipeline source.

    Returns:
        The pipeline.
    """
    try:
        pipeline_instance = source_utils.load(source)
    except ModuleNotFoundError as e:
        source_root = source_utils.get_source_root()
        cli_utils.error(
            f"Unable to import module `{e.name}`. Make sure the source path is "
            f"relative to your source root `{source_root}`."
        )
    except AttributeError as e:
        cli_utils.error("Unable to load attribute from module: " + str(e))

    if not isinstance(pipeline_instance, Pipeline):
        cli_utils.error(
            f"The given source path `{source}` does not resolve to a pipeline "
            "object."
        )

    return pipeline_instance


@cli.group(cls=TagGroup, tag=CliCategories.MANAGEMENT_TOOLS)
def deployment() -> None:
    """Interact with deployments."""


@deployment.command("list", help="List all registered deployments.")
@list_options(
    DeploymentFilter,
    default_columns=[
        "id",
        "name",
        "status",
        "url",
        "pipeline",
        "stack",
    ],
)
def list_deployments(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all registered deployments for the filter.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter deployments.
    """
    client = Client()
    try:
        with console.status("Listing deployments...\n"):
            deployments = client.list_deployments(**kwargs)
    except KeyError as err:
        cli_utils.exception(err)
        return

    cli_utils.print_page(
        deployments,
        columns,
        output_format,
        cli_utils.generate_deployment_row,
        empty_message="No deployments found for this filter.",
    )


@deployment.command("describe", help="Describe a deployment.")
@click.argument("deployment_name_or_id", type=str, required=True)
@click.option(
    "--show-secret",
    "-s",
    is_flag=True,
    help="Show the secret key.",
)
@click.option(
    "--show-metadata",
    "-m",
    is_flag=True,
    help="Show the metadata.",
)
@click.option(
    "--show-schema",
    "-sc",
    is_flag=True,
    help="Show the schema.",
)
@click.option(
    "--no-truncate",
    "-nt",
    is_flag=True,
    help="Don't truncate the metadata.",
)
def describe_deployment(
    deployment_name_or_id: str,
    show_secret: bool = False,
    show_metadata: bool = False,
    no_truncate: bool = False,
    show_schema: bool = False,
) -> None:
    """Describe a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to describe.
        show_secret: If True, show the secret key.
        show_metadata: If True, show the metadata.
        show_schema: If True, show the schema.
        no_truncate: If True, don't truncate the metadata.
    """
    try:
        deployment = Client().get_deployment(
            name_id_or_prefix=deployment_name_or_id,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.pretty_print_deployment(
            deployment,
            show_secret=show_secret,
            show_metadata=show_metadata,
            show_schema=show_schema,
            no_truncate=no_truncate,
        )


@deployment.command("provision", help="Provision a deployment.")
@click.argument("deployment_name_or_id", type=str, required=True)
@click.option(
    "--snapshot",
    "-s",
    "snapshot_name_or_id",
    type=str,
    required=False,
    help="ID or name of the snapshot to use. If not provided, the current "
    "snapshot configured for the deployment will be used.",
)
@click.option(
    "--pipeline",
    "-p",
    "pipeline_name_or_id",
    type=str,
    required=False,
    help="The name or ID of the pipeline to which the snapshot belongs.",
)
@click.option(
    "--overtake",
    "-o",
    "overtake",
    is_flag=True,
    default=False,
    required=False,
    help="Provision the deployment with the given name even if it is "
    "owned by a different user.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the deployment to be "
    "provisioned.",
)
def provision_deployment(
    deployment_name_or_id: str,
    snapshot_name_or_id: Optional[str] = None,
    pipeline_name_or_id: Optional[str] = None,
    overtake: bool = False,
    timeout: Optional[int] = None,
) -> None:
    """Provision a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to deploy.
        snapshot_name_or_id: The ID or name of the pipeline snapshot to use.
        pipeline_name_or_id: The name or ID of the pipeline to which the
            snapshot belongs.
        overtake: If True, provision the deployment with the given name
            even if it is owned by a different user.
        timeout: The maximum time in seconds to wait for the deployment
            to be provisioned.
    """
    snapshot_id: Optional[UUID] = None
    if snapshot_name_or_id:
        snapshot = fetch_snapshot(snapshot_name_or_id, pipeline_name_or_id)
        snapshot_id = snapshot.id

    client = Client()
    try:
        deployment = client.get_deployment(deployment_name_or_id)
    except KeyError:
        pass
    else:
        if (
            deployment.user
            and deployment.user.id != client.active_user.id
            and not overtake
        ):
            confirmation = cli_utils.confirmation(
                f"Deployment with name '{deployment.name}' is owned by a "
                f"different user '{deployment.user.name}'.\nDo you want to "
                "continue and provision it "
                "(hint: use the --overtake flag to skip this check)?"
            )
            if not confirmation:
                cli_utils.declare("Deployment provisioning canceled.")
                return

    with console.status(
        f"Provisioning deployment '{deployment_name_or_id}'...\n"
    ):
        try:
            deployment = Client().provision_deployment(
                name_id_or_prefix=deployment_name_or_id,
                snapshot_id=snapshot_id,
                timeout=timeout,
            )
        except KeyError as e:
            cli_utils.exception(e)
        else:
            cli_utils.declare(
                f"Provisioned deployment '{deployment_name_or_id}'."
            )
            cli_utils.pretty_print_deployment(deployment, show_secret=True)

            dashboard_url = get_deployment_url(deployment)
            if dashboard_url:
                cli_utils.declare(
                    f"\nView in ZenML Cloud: [link]{dashboard_url}[/link]",
                    no_wrap=True,
                )


@deployment.command("deprovision", help="Deprovision a deployment.")
@click.argument("deployment_name_or_id", type=str, required=False)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Deprovision all deployments.",
)
@click.option(
    "--mine",
    "-m",
    is_flag=True,
    default=False,
    help="Deprovision only deployments owned by the current user.",
)
@click.option(
    "--ignore-errors",
    "-i",
    is_flag=True,
    default=False,
    help="Ignore errors when deprovisioning multiple deployments.",
)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation.",
)
@click.option(
    "--max-count",
    "-c",
    "max_count",
    type=int,
    required=False,
    default=10,
    help="Maximum number of deployments to deprovision in one go.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the deployment to be "
    "deprovisioned.",
)
def deprovision_deployment(
    deployment_name_or_id: Optional[str] = None,
    all: bool = False,
    mine: bool = False,
    yes: bool = False,
    ignore_errors: bool = False,
    max_count: int = 10,
    timeout: Optional[int] = None,
) -> None:
    """Deprovision a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to deprovision.
        all: If set, deprovision all deployments.
        mine: If set, deprovision only deployments owned by the current user.
        yes: If set, don't ask for confirmation.
        ignore_errors: If set, ignore errors when deprovisioning multiple
            deployments.
        max_count: The maximum number of deployments to deprovision in one go.
        timeout: The maximum time in seconds to wait for the deployment
            to be deprovisioned.
    """
    client = Client()

    if all:
        deployments = client.list_deployments(size=max_count).items
    elif mine:
        deployments = client.list_deployments(
            user=client.active_user.id,
            size=max_count,
        ).items
    elif deployment_name_or_id:
        deployments = [
            client.get_deployment(name_id_or_prefix=deployment_name_or_id)
        ]
    else:
        cli_utils.error(
            "Either a deployment name or ID must be provided or --all or "
            "--mine must be set."
        )

    if len(deployments) == 0:
        cli_utils.error("No deployments found to deprovision.")

    if len(deployments) == 1:
        cli_utils.declare(
            f"The following deployment will be deprovisioned: "
            f"{deployments[0].name}"
        )
    else:
        deployment_names = [deployment.name for deployment in deployments]
        cli_utils.declare(
            f"The following deployments will ALL be deprovisioned: "
            f"{', '.join(deployment_names)}"
        )

    other_deployments = [
        deployment
        for deployment in deployments
        if deployment.user and deployment.user.id != client.active_user.id
    ]

    if other_deployments:
        deployment_names = [
            deployment.name for deployment in other_deployments
        ]
        cli_utils.warning(
            f"The following deployments are not owned by the current "
            f"user: {', '.join(deployment_names)} !"
        )

    if not yes:
        confirmation = cli_utils.confirmation(
            "Are you sure you want to continue?"
        )
        if not confirmation:
            cli_utils.declare("Deployment deprovision canceled.")
            return

    for deployment in deployments:
        with console.status(
            f"Deprovisioning deployment '{deployment.name}'...\n"
        ):
            try:
                client.deprovision_deployment(
                    name_id_or_prefix=deployment.id,
                    timeout=timeout,
                )
                cli_utils.declare(
                    f"Deprovisioned deployment '{deployment.name}'."
                )
                cli_utils.declare(
                    "Hint: to permanently delete the deployment, run `zenml "
                    f"deployment delete {deployment.name}`."
                )
            except KeyError as e:
                error_message = (
                    f"Failed to deprovision deployment '{deployment.name}': "
                    f"{str(e)}"
                )
                if all or mine and ignore_errors:
                    cli_utils.warning(error_message)
                else:
                    cli_utils.error(error_message)


@deployment.command("delete", help="Delete a deployment.")
@click.argument("deployment_name_or_id", type=str, required=False)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Deprovision all deployments.",
)
@click.option(
    "--mine",
    "-m",
    is_flag=True,
    default=False,
    help="Deprovision only deployments owned by the current user.",
)
@click.option(
    "--ignore-errors",
    "-i",
    is_flag=True,
    default=False,
    help="Ignore errors when deprovisioning multiple deployments.",
)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the deployment to be "
    "deprovisioned.",
)
@click.option(
    "--max-count",
    "-c",
    "max_count",
    type=int,
    required=False,
    default=20,
    help="Maximum number of deployments to delete in one go.",
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    default=False,
    help="Force the deletion of the deployment if it cannot be deprovisioned.",
)
def delete_deployment(
    deployment_name_or_id: Optional[str] = None,
    all: bool = False,
    mine: bool = False,
    ignore_errors: bool = False,
    yes: bool = False,
    timeout: Optional[int] = None,
    max_count: int = 20,
    force: bool = False,
) -> None:
    """Deprovision and delete a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to delete.
        all: If set, delete all deployments.
        mine: If set, delete only deployments owned by the current user.
        ignore_errors: If set, ignore errors when deleting multiple deployments.
        yes: If set, don't ask for confirmation.
        timeout: The maximum time in seconds to wait for the deployment
            to be deprovisioned.
        max_count: The maximum number of deployments to delete in one go.
        force: If set, force the deletion of the deployment if it cannot
            be deprovisioned.
    """
    client = Client()

    if all:
        deployments = client.list_deployments(size=max_count).items
    elif mine:
        deployments = client.list_deployments(
            user=client.active_user.id,
            size=max_count,
        ).items
    elif deployment_name_or_id:
        deployments = [
            client.get_deployment(name_id_or_prefix=deployment_name_or_id)
        ]
    else:
        cli_utils.error(
            "Either a deployment name or ID must be provided or --all or "
            "--mine must be set."
        )

    if len(deployments) == 0:
        cli_utils.error("No deployments found to delete.")

    if len(deployments) == 1:
        cli_utils.declare(
            f"The following deployment will be deleted: {deployments[0].name}"
        )
    else:
        deployment_names = [deployment.name for deployment in deployments]
        cli_utils.declare(
            f"The following deployments will ALL be deleted: "
            f"{', '.join(deployment_names)}"
        )

    other_deployments = [
        deployment
        for deployment in deployments
        if deployment.user and deployment.user.id != client.active_user.id
    ]

    if other_deployments:
        deployment_names = [
            deployment.name for deployment in other_deployments
        ]
        cli_utils.warning(
            f"The following deployments are not owned by the current "
            f"user: {', '.join(deployment_names)} !"
        )

    if not yes:
        confirmation = cli_utils.confirmation(
            "Are you sure you want to continue?"
        )
        if not confirmation:
            cli_utils.declare("Deployment deletion canceled.")
            return

    for deployment in deployments:
        with console.status(f"Deleting deployment '{deployment.name}'...\n"):
            try:
                Client().delete_deployment(
                    name_id_or_prefix=deployment.id,
                    force=force,
                    timeout=timeout,
                )
                cli_utils.declare(f"Deleted deployment '{deployment.name}'.")
            except KeyError as e:
                error_message = (
                    f"Failed to delete deployment '{deployment.name}': "
                    f"{str(e)}"
                )
                if all or mine and ignore_errors:
                    cli_utils.warning(error_message)
                else:
                    cli_utils.error(error_message)


@deployment.command("refresh", help="Refresh the status of a deployment.")
@click.argument("deployment_name_or_id", type=str, required=True)
def refresh_deployment(
    deployment_name_or_id: str,
) -> None:
    """Refresh the status of a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to refresh.
    """
    try:
        deployment = Client().refresh_deployment(
            name_id_or_prefix=deployment_name_or_id
        )

    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.pretty_print_deployment(deployment, show_secret=True)


@deployment.command(
    "invoke",
    context_settings={"ignore_unknown_options": True},
    help="Invoke a deployment with arguments.",
)
@click.argument("deployment_name_or_id", type=str, required=True)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the deployment to be invoked.",
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def invoke_deployment(
    deployment_name_or_id: str,
    args: List[str],
    timeout: Optional[int] = None,
) -> None:
    """Call a deployment with arguments.

    Args:
        deployment_name_or_id: The name or ID of the deployment to call.
        args: The arguments to pass to the deployment call.
        timeout: The maximum time in seconds to wait for the deployment
            to be invoked.
    """
    from zenml.deployers.utils import invoke_deployment

    # Parse the given args
    args = list(args)
    args.append(deployment_name_or_id)

    name_or_id, parsed_args = cli_utils.parse_name_and_extra_arguments(
        args,
        expand_args=True,
        name_mandatory=True,
    )
    assert name_or_id is not None

    try:
        response = invoke_deployment(
            deployment_name_or_id=name_or_id,
            timeout=timeout or 300,  # 5 minute timeout
            project=None,
            **parsed_args,
        )
    except DeploymentInvalidParametersError as e:
        cli_utils.error(
            f"Invalid parameters for deployment '{name_or_id}': \n"
            f"{str(e)}\n\n"
            f"Hint: run 'zenml deployment describe --schema {name_or_id}' "
            "to inspect the deployment schema."
        )
    except KeyError as e:
        cli_utils.error(
            str(e)
            + "\n"
            + f"Hint: run [green]`zenml deployment logs {name_or_id}`[/green] "
            "to inspect the deployment logs."
        )
    else:
        cli_utils.declare(f"Invoked deployment '{name_or_id}' with response:")
        print(json.dumps(response, indent=2))
        if isinstance(response, dict) and not response.get("success", True):
            cli_utils.declare(
                f"Hint: run [green]`zenml deployment logs {name_or_id}`[/green] "
                "to inspect the deployment logs."
            )


@deployment.command("logs", help="Get the logs of a deployment.")
@click.argument("deployment_name_or_id", type=str, required=True)
@click.option(
    "--follow",
    "-f",
    is_flag=True,
    default=False,
    help="Follow the logs.",
)
@click.option(
    "--tail",
    "-t",
    type=int,
    default=None,
    help="The number of lines to show from the end of the logs.",
)
def log_deployment(
    deployment_name_or_id: str,
    follow: bool = False,
    tail: Optional[int] = None,
) -> None:
    """Get the logs of a deployment.

    Args:
        deployment_name_or_id: The name or ID of the deployment to get the logs of.
        follow: If True, follow the logs.
        tail: The number of lines to show from the end of the logs. If None,
            show all logs.
    """
    try:
        logs = Client().get_deployment_logs(
            name_id_or_prefix=deployment_name_or_id,
            follow=follow,
            tail=tail,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        for log in logs:
            print(log)



# Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""SQLModel implementation of curated visualization tables."""

from typing import TYPE_CHECKING, Any, List, Optional, Sequence
from uuid import UUID

from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import selectinload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship

from zenml.enums import CuratedVisualizationSize, VisualizationResourceTypes
from zenml.models.v2.core.curated_visualization import (
    CuratedVisualizationRequest,
    CuratedVisualizationResponse,
    CuratedVisualizationResponseBody,
    CuratedVisualizationResponseMetadata,
    CuratedVisualizationResponseResources,
    CuratedVisualizationUpdate,
)
from zenml.zen_stores.schemas.base_schemas import BaseSchema
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
)
from zenml.zen_stores.schemas.utils import jl_arg

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.artifact_visualization_schemas import (
        ArtifactVisualizationSchema,
    )


class CuratedVisualizationSchema(BaseSchema, table=True):
    """SQL Model for curated visualizations."""

    __tablename__ = "curated_visualization"
    __table_args__ = (
        UniqueConstraint(
            "artifact_visualization_id",
            "resource_id",
            "resource_type",
            name="unique_curated_visualization_resource_link",
        ),
    )

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    artifact_visualization_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target="artifact_visualization",
        source_column="artifact_visualization_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        custom_constraint_name="fk_curated_visualization_artifact_visualization_id",
    )

    display_name: Optional[str] = Field(default=None)
    display_order: Optional[int] = Field(default=None)
    layout_size: str = Field(
        default=CuratedVisualizationSize.FULL_WIDTH.value,
        nullable=False,
    )
    resource_id: UUID = Field(nullable=False)
    resource_type: str = Field(nullable=False)

    artifact_visualization: "ArtifactVisualizationSchema" = Relationship(
        back_populates="curated_visualizations"
    )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options: List[ExecutableOption] = []

        if include_resources:
            options.append(selectinload(jl_arg(cls.artifact_visualization)))

        return options

    @classmethod
    def from_request(
        cls, request: CuratedVisualizationRequest
    ) -> "CuratedVisualizationSchema":
        """Convert a request into a schema instance.

        Args:
            request: The request to convert.

        Returns:
            The created schema.
        """
        return cls(
            project_id=request.project,
            artifact_visualization_id=request.artifact_visualization_id,
            display_name=request.display_name,
            display_order=request.display_order,
            layout_size=request.layout_size.value,
            resource_id=request.resource_id,
            resource_type=request.resource_type.value,
        )

    def update(
        self,
        update: CuratedVisualizationUpdate,
    ) -> "CuratedVisualizationSchema":
        """Update a schema instance from an update model.

        Args:
            update: The update definition.

        Returns:
            The updated schema.
        """
        changes = update.model_dump(exclude_unset=True)
        layout_size_update = changes.pop("layout_size", None)
        if layout_size_update is not None:
            self.layout_size = layout_size_update.value

        for field, value in changes.items():
            if hasattr(self, field):
                setattr(self, field, value)

        from zenml.utils.time_utils import utc_now

        self.updated = utc_now()
        return self

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> CuratedVisualizationResponse:
        """Convert schema into response model.

        Args:
            include_metadata: Whether to include metadata in the response.
            include_resources: Whether to include resources in the response.
            **kwargs: Additional keyword arguments.

        Returns:
            The created response model.
        """
        try:
            layout_size_enum = CuratedVisualizationSize(self.layout_size)
        except ValueError:
            layout_size_enum = CuratedVisualizationSize.FULL_WIDTH

        try:
            resource_type_enum = VisualizationResourceTypes(self.resource_type)
        except ValueError:
            resource_type_enum = VisualizationResourceTypes.PROJECT

        artifact_version_id = self.artifact_visualization.artifact_version_id

        body = CuratedVisualizationResponseBody(
            project_id=self.project_id,
            created=self.created,
            updated=self.updated,
            artifact_visualization_id=self.artifact_visualization_id,
            artifact_version_id=artifact_version_id,
            display_name=self.display_name,
            display_order=self.display_order,
            layout_size=layout_size_enum,
            resource_id=self.resource_id,
            resource_type=resource_type_enum,
        )

        metadata = None
        if include_metadata:
            metadata = CuratedVisualizationResponseMetadata()

        resources = None
        if include_resources:
            artifact_visualization = self.artifact_visualization.to_model(
                include_metadata=False,
                include_resources=False,
            )
            resources = CuratedVisualizationResponseResources(
                artifact_visualization=artifact_visualization,
            )

        return CuratedVisualizationResponse(
            id=self.id,
            body=body,
            metadata=metadata,
            resources=resources,
        )



"""Release [0.91.2].

Revision ID: 0.91.2
Revises: 4dd9d3afd2c0
Create Date: 2025-11-19 00:54:27.184728

"""

# revision identifiers, used by Alembic.
revision = "0.91.2"
down_revision = "4dd9d3afd2c0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



"""Split up step configurations [3d7e39f3ac92].

Revision ID: 3d7e39f3ac92
Revises: 857843db1bcf
Create Date: 2025-06-17 17:45:31.702617

"""

import json
import uuid

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.dialects import mysql

from zenml.utils.time_utils import utc_now

# revision identifiers, used by Alembic.
revision = "3d7e39f3ac92"
down_revision = "857843db1bcf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "step_configuration",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("index", sa.Integer(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column(
            "config",
            sa.String(length=16777215).with_variant(mysql.MEDIUMTEXT, "mysql"),
            nullable=False,
        ),
        sa.Column(
            "deployment_id", sqlmodel.sql.sqltypes.GUID(), nullable=False
        ),
        sa.ForeignKeyConstraint(
            ["deployment_id"],
            ["pipeline_deployment.id"],
            name="fk_step_configuration_deployment_id_pipeline_deployment",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "deployment_id", "name", name="unique_step_name_for_deployment"
        ),
    )
    with op.batch_alter_table("pipeline_deployment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("step_count", sa.Integer(), nullable=True)
        )

    # Migrate existing step configurations
    connection = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(
        bind=connection, only=("pipeline_deployment", "step_configuration")
    )
    pipeline_deployment_table = sa.Table("pipeline_deployment", meta)
    step_configuration_table = sa.Table("step_configuration", meta)

    step_configurations_to_insert = []
    deployment_updates = []

    for deployment_id, step_configurations_json in connection.execute(
        sa.select(
            pipeline_deployment_table.c.id,
            pipeline_deployment_table.c.step_configurations,
        )
    ):
        step_configurations = json.loads(step_configurations_json)

        step_count = len(step_configurations)
        deployment_updates.append(
            {
                "id_": deployment_id,
                "step_count": step_count,
            }
        )

        for index, (step_name, step_config) in enumerate(
            step_configurations.items()
        ):
            now = utc_now()
            step_configurations_to_insert.append(
                {
                    "id": str(uuid.uuid4()).replace("-", ""),
                    "created": now,
                    "updated": now,
                    "index": index,
                    "name": step_name,
                    "config": json.dumps(step_config),
                    "deployment_id": deployment_id,
                }
            )

    op.bulk_insert(
        step_configuration_table, rows=step_configurations_to_insert
    )
    if deployment_updates:
        connection.execute(
            sa.update(pipeline_deployment_table).where(
                pipeline_deployment_table.c.id == sa.bindparam("id_")
            ),
            deployment_updates,
        )

    with op.batch_alter_table("pipeline_deployment", schema=None) as batch_op:
        batch_op.alter_column(
            "step_count", existing_type=sa.Integer(), nullable=False
        )
        batch_op.drop_column("step_configurations")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_deployment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "step_configurations",
                sa.VARCHAR(length=16777215),
                nullable=False,
            )
        )
        batch_op.drop_column("step_count")

    op.drop_table("step_configuration")
    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Runner entrypoint configuration."""

from typing import Any, Dict, List
from uuid import UUID

from zenml.client import Client
from zenml.constants import is_true_string_value
from zenml.entrypoints.base_entrypoint_configuration import (
    BaseEntrypointConfiguration,
)
from zenml.execution.pipeline.utils import submit_pipeline

PLACEHOLDER_RUN_ID_OPTION = "placeholder_run_id"
RUN_ID_OPTION = "run_id"
RESUME_OPTION = "resume"


class RunnerEntrypointConfiguration(BaseEntrypointConfiguration):
    """Runner entrypoint configuration."""

    @classmethod
    def get_entrypoint_options(cls) -> Dict[str, bool]:
        """Gets all options required for running with this configuration.

        Returns:
            The superclass options as well as options for the run ID and
            whether the run should be resumed.
        """
        return super().get_entrypoint_options() | {
            RUN_ID_OPTION: True,
            RESUME_OPTION: False,
        }

    @classmethod
    def get_entrypoint_arguments(
        cls,
        **kwargs: Any,
    ) -> List[str]:
        """Gets all arguments that the entrypoint command should be called with.

        Args:
            **kwargs: Kwargs, must include the run ID. The optional `resume`
                flag enables the resume path.

        Returns:
            The superclass arguments as well as arguments for the run ID and
            optional resume flag.
        """
        arguments = super().get_entrypoint_arguments(**kwargs) + [
            f"--{RUN_ID_OPTION}",
            str(kwargs[RUN_ID_OPTION]),
            f"--{PLACEHOLDER_RUN_ID_OPTION}",
            str(kwargs[RUN_ID_OPTION]),
        ]
        if RESUME_OPTION in kwargs:
            arguments.extend(
                [f"--{RESUME_OPTION}", str(kwargs[RESUME_OPTION])]
            )
        return arguments

    def run(self) -> None:
        """Run the entrypoint configuration.

        This method runs or resumes the pipeline defined by the configured
        snapshot.
        """
        snapshot = self.snapshot
        run = Client().get_pipeline_run(
            UUID(self.entrypoint_args[RUN_ID_OPTION])
        )

        stack = Client().active_stack
        assert snapshot.stack and stack.id == snapshot.stack.id

        resume = is_true_string_value(self.entrypoint_args.get(RESUME_OPTION))
        if resume:
            stack.orchestrator.resume_run(
                snapshot=snapshot, run=run, stack=stack, force_async=True
            )
        else:
            submit_pipeline(
                snapshot=snapshot,
                stack=stack,
                placeholder_run=run,
            )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing Pipeline Deployments."""

from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from pydantic import BaseModel, Field

from zenml.constants import STR_FIELD_MAX_LENGTH
from zenml.enums import DeploymentStatus
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.scoped import (
    ProjectScopedFilter,
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
    TaggableFilter,
)
from zenml.utils.tag_utils import Tag

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.models.v2.core.component import ComponentResponse
    from zenml.models.v2.core.curated_visualization import (
        CuratedVisualizationResponse,
    )
    from zenml.models.v2.core.pipeline import PipelineResponse
    from zenml.models.v2.core.pipeline_snapshot import (
        PipelineSnapshotResponse,
    )
    from zenml.models.v2.core.stack import StackResponse
    from zenml.models.v2.core.tag import TagResponse
    from zenml.zen_stores.schemas.base_schemas import BaseSchema

    AnySchema = TypeVar("AnySchema", bound=BaseSchema)
    AnyQuery = TypeVar("AnyQuery", bound=Any)


class DeploymentOperationalState(BaseModel):
    """Operational state of a deployment."""

    status: DeploymentStatus = Field(default=DeploymentStatus.UNKNOWN)
    url: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None


# ------------------ Request Model ------------------


class DeploymentRequest(ProjectScopedRequest):
    """Request model for deployments."""

    name: str = Field(
        title="The name of the deployment.",
        description="A unique name for the deployment within the project.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    snapshot_id: UUID = Field(
        title="The pipeline snapshot ID.",
        description="The ID of the pipeline snapshot associated with the "
        "deployment.",
    )
    deployer_id: UUID = Field(
        title="The deployer ID.",
        description="The ID of the deployer component managing this deployment.",
    )
    auth_key: Optional[str] = Field(
        default=None,
        title="The auth key of the deployment.",
        description="The auth key of the deployment.",
    )
    tags: Optional[List[Union[str, Tag]]] = Field(
        default=None,
        title="Tags of the deployment.",
    )


# ------------------ Update Model ------------------


class DeploymentUpdate(BaseUpdate):
    """Update model for deployments."""

    name: Optional[str] = Field(
        default=None,
        title="The new name of the deployment.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    snapshot_id: Optional[UUID] = Field(
        default=None,
        title="New pipeline snapshot ID.",
    )
    url: Optional[str] = Field(
        default=None,
        title="The new URL of the deployment.",
    )
    status: Optional[DeploymentStatus] = Field(
        default=None,
        title="The new status of the deployment.",
    )
    deployment_metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        title="The new metadata of the deployment.",
    )
    auth_key: Optional[str] = Field(
        default=None,
        title="The new auth key of the deployment.",
    )
    add_tags: Optional[List[str]] = Field(
        default=None, title="New tags to add to the deployment."
    )
    remove_tags: Optional[List[str]] = Field(
        default=None, title="Tags to remove from the deployment."
    )

    @classmethod
    def from_operational_state(
        cls, operational_state: DeploymentOperationalState
    ) -> "DeploymentUpdate":
        """Create an update from an operational state.

        Args:
            operational_state: The operational state to create an update from.

        Returns:
            The update.
        """
        return cls(
            status=operational_state.status,
            url=operational_state.url,
            deployment_metadata=operational_state.metadata,
        )


# ------------------ Response Model ------------------


class DeploymentResponseBody(ProjectScopedResponseBody):
    """Response body for deployments."""

    url: Optional[str] = Field(
        default=None,
        title="The URL of the deployment.",
        description="The HTTP URL where the deployment can be accessed.",
    )
    status: Optional[DeploymentStatus] = Field(
        default=None,
        title="The status of the deployment.",
        description="Current operational status of the deployment.",
    )


class DeploymentResponseMetadata(ProjectScopedResponseMetadata):
    """Response metadata for deployments."""

    deployment_metadata: Dict[str, Any] = Field(
        title="The metadata of the deployment.",
    )
    auth_key: Optional[str] = Field(
        default=None,
        title="The auth key of the deployment.",
        description="The auth key of the deployment.",
    )


class DeploymentResponseResources(ProjectScopedResponseResources):
    """Response resources for deployments."""

    snapshot: Optional["PipelineSnapshotResponse"] = Field(
        default=None,
        title="The pipeline snapshot.",
        description="The pipeline snapshot being deployed.",
    )
    deployer: Optional["ComponentResponse"] = Field(
        default=None,
        title="The deployer.",
        description="The deployer component managing this deployment.",
    )
    pipeline: Optional["PipelineResponse"] = Field(
        default=None,
        title="The pipeline.",
        description="The pipeline being deployed.",
    )
    tags: List["TagResponse"] = Field(
        title="Tags associated with the deployment.",
    )
    visualizations: List["CuratedVisualizationResponse"] = Field(
        default_factory=list,
        title="Curated deployment visualizations.",
    )
    stack: Optional["StackResponse"] = Field(
        default=None,
        title="The stack that was deployed.",
        description="The stack that was deployed.",
    )


class DeploymentResponse(
    ProjectScopedResponse[
        DeploymentResponseBody,
        DeploymentResponseMetadata,
        DeploymentResponseResources,
    ]
):
    """Response model for deployments."""

    name: str = Field(
        title="The name of the deployment.",
        max_length=STR_FIELD_MAX_LENGTH,
    )

    def get_hydrated_version(self) -> "DeploymentResponse":
        """Get the hydrated version of this deployment.

        Returns:
            an instance of the same entity with the metadata and resources fields
            attached.
        """
        from zenml.client import Client

        client = Client()
        return client.get_deployment(self.id)

    # Helper properties
    @property
    def url(self) -> Optional[str]:
        """The URL of the deployment.

        Returns:
            The URL of the deployment.
        """
        return self.get_body().url

    @property
    def status(self) -> Optional[DeploymentStatus]:
        """The status of the deployment.

        Returns:
            The status of the deployment.
        """
        return self.get_body().status

    @property
    def deployment_metadata(self) -> Dict[str, Any]:
        """The metadata of the deployment.

        Returns:
            The metadata of the deployment.
        """
        return self.get_metadata().deployment_metadata

    @property
    def auth_key(self) -> Optional[str]:
        """The auth key of the deployment.

        Returns:
            The auth key of the deployment.
        """
        return self.get_metadata().auth_key

    @property
    def snapshot(self) -> Optional["PipelineSnapshotResponse"]:
        """The pipeline snapshot.

        Returns:
            The pipeline snapshot.
        """
        return self.get_resources().snapshot

    @property
    def deployer(self) -> Optional["ComponentResponse"]:
        """The deployer.

        Returns:
            The deployer.
        """
        return self.get_resources().deployer

    @property
    def pipeline(self) -> Optional["PipelineResponse"]:
        """The pipeline.

        Returns:
            The pipeline.
        """
        return self.get_resources().pipeline

    def tags(self) -> List["TagResponse"]:
        """The tags of the deployment.

        Returns:
            The tags of the deployment.
        """
        return self.get_resources().tags

    @property
    def visualizations(self) -> List["CuratedVisualizationResponse"]:
        """The visualizations of the deployment.

        Returns:
            The visualizations of the deployment.
        """
        return self.get_resources().visualizations

    @property
    def stack(self) -> Optional["StackResponse"]:
        """The stack.

        Returns:
            The stack.
        """
        return self.get_resources().stack

    @property
    def snapshot_id(self) -> Optional[UUID]:
        """The pipeline snapshot ID.

        Returns:
            The pipeline snapshot ID.
        """
        snapshot = self.get_resources().snapshot
        if snapshot:
            return snapshot.id
        return None

    @property
    def deployer_id(self) -> Optional[UUID]:
        """The deployer ID.

        Returns:
            The deployer ID.
        """
        deployer = self.get_resources().deployer
        if deployer:
            return deployer.id
        return None


# ------------------ Filter Model ------------------


class DeploymentFilter(ProjectScopedFilter, TaggableFilter):
    """Model to enable advanced filtering of deployments."""

    CUSTOM_SORTING_OPTIONS: ClassVar[List[str]] = [
        *ProjectScopedFilter.CUSTOM_SORTING_OPTIONS,
        *TaggableFilter.CUSTOM_SORTING_OPTIONS,
        "snapshot",
        "pipeline",
    ]
    FILTER_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *ProjectScopedFilter.FILTER_EXCLUDE_FIELDS,
        *TaggableFilter.FILTER_EXCLUDE_FIELDS,
        "pipeline",
    ]
    CLI_EXCLUDE_FIELDS = [
        *ProjectScopedFilter.CLI_EXCLUDE_FIELDS,
        *TaggableFilter.CLI_EXCLUDE_FIELDS,
    ]

    name: Optional[str] = Field(
        default=None,
        description="Name of the deployment.",
    )
    url: Optional[str] = Field(
        default=None,
        description="URL of the deployment.",
    )
    status: Optional[str] = Field(
        default=None,
        description="Status of the deployment.",
    )
    pipeline: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Pipeline associated with the deployment.",
        union_mode="left_to_right",
    )
    snapshot_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Pipeline snapshot ID associated with the deployment.",
        union_mode="left_to_right",
    )
    deployer_id: Optional[Union[UUID, str]] = Field(
        default=None,
        description="Deployer ID managing the deployment.",
        union_mode="left_to_right",
    )

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        from sqlmodel import and_

        from zenml.zen_stores.schemas import (
            DeploymentSchema,
            PipelineSchema,
            PipelineSnapshotSchema,
        )

        custom_filters = super().get_custom_filters(table)

        if self.pipeline:
            pipeline_filter = and_(
                DeploymentSchema.snapshot_id == PipelineSnapshotSchema.id,
                PipelineSnapshotSchema.pipeline_id == PipelineSchema.id,
                self.generate_name_or_id_query_conditions(
                    value=self.pipeline, table=PipelineSchema
                ),
            )
            custom_filters.append(pipeline_filter)

        return custom_filters

    def apply_sorting(
        self,
        query: "AnyQuery",
        table: Type["AnySchema"],
    ) -> "AnyQuery":
        """Apply sorting to the query.

        Args:
            query: The query to which to apply the sorting.
            table: The query table.

        Returns:
            The query with sorting applied.
        """
        from sqlmodel import asc, desc

        from zenml.enums import SorterOps
        from zenml.zen_stores.schemas import (
            DeploymentSchema,
            PipelineSchema,
            PipelineSnapshotSchema,
        )

        sort_by, operand = self.sorting_params

        if sort_by == "pipeline":
            query = query.outerjoin(
                PipelineSnapshotSchema,
                DeploymentSchema.snapshot_id == PipelineSnapshotSchema.id,
            ).outerjoin(
                PipelineSchema,
                PipelineSnapshotSchema.pipeline_id == PipelineSchema.id,
            )
            column: Any = PipelineSchema.name
        elif sort_by == "snapshot":
            query = query.outerjoin(
                PipelineSnapshotSchema,
                DeploymentSchema.snapshot_id == PipelineSnapshotSchema.id,
            )
            column = PipelineSnapshotSchema.name
        else:
            return super().apply_sorting(query=query, table=table)

        query = query.add_columns(column)

        if operand == SorterOps.ASCENDING:
            query = query.order_by(asc(column))
        else:
            query = query.order_by(desc(column))

        return query



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for linking step outputs to model versions."""

from typing import Dict, Optional, Union
from uuid import UUID

from zenml.client import Client
from zenml.enums import ModelStages
from zenml.exceptions import StepContextError
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType
from zenml.model.model import Model
from zenml.models import (
    ArtifactVersionResponse,
    ModelVersionArtifactRequest,
    ModelVersionResponse,
    ServiceUpdate,
)
from zenml.steps.step_context import get_step_context

logger = get_logger(__name__)


def log_model_metadata(
    metadata: Dict[str, "MetadataType"],
    model_name: Optional[str] = None,
    model_version: Optional[Union[ModelStages, int, str]] = None,
) -> None:
    """Log model version metadata.

    This function can be used to log metadata for existing model versions.

    Args:
        metadata: The metadata to log.
        model_name: The name of the model to log metadata for. Can
            be omitted when being called inside a step with configured
            `model` in decorator.
        model_version: The version of the model to log metadata for. Can
            be omitted when being called inside a step with configured
            `model` in decorator.

    Raises:
        ValueError: If the function is not called with proper input.
    """
    logger.warning(
        "The `log_model_metadata` function is deprecated and will soon be "
        "removed. Instead, you can consider using: "
        "`log_metadata(metadata={...}, infer_model=True)` instead. For more "
        "info: https://docs.zenml.io/concepts/metadata#attaching-metadata-to-models"
    )

    from zenml import log_metadata

    if model_name and model_version:
        log_metadata(
            metadata=metadata,
            model_version=model_version,
            model_name=model_name,
        )
    elif model_name is None and model_version is None:
        log_metadata(
            metadata=metadata,
            infer_model=True,
        )
    else:
        raise ValueError(
            "You can call `log_model_metadata` by either providing both "
            "`model_name` and `model_version` or keeping both of them None."
        )


def link_artifact_version_to_model_version(
    artifact_version: ArtifactVersionResponse,
    model_version: ModelVersionResponse,
) -> None:
    """Link an artifact version to a model version.

    Args:
        artifact_version: The artifact version to link.
        model_version: The model version to link.
    """
    client = Client()
    client.zen_store.create_model_version_artifact_link(
        ModelVersionArtifactRequest(
            artifact_version=artifact_version.id,
            model_version=model_version.id,
        )
    )


def link_artifact_to_model(
    artifact_version: ArtifactVersionResponse,
    model: Optional["Model"] = None,
) -> None:
    """Link the artifact to the model.

    Args:
        artifact_version: The artifact version to link.
        model: The model to link to.

    Raises:
        RuntimeError: If called outside a step.
    """
    if not model:
        is_issue = False
        try:
            step_context = get_step_context()
            model = step_context.model
        except StepContextError:
            is_issue = True

        if model is None or is_issue:
            raise RuntimeError(
                "`link_artifact_to_model` called without `model` parameter "
                "and configured model context cannot be identified. Consider "
                "passing the `model` explicitly or configuring it in "
                "@step or @pipeline decorator."
            )

    model_version = model._get_or_create_model_version()
    link_artifact_version_to_model_version(
        artifact_version=artifact_version,
        model_version=model_version,
    )


def link_service_to_model(
    service_id: UUID,
    model: Optional["Model"] = None,
    model_version_id: Optional[UUID] = None,
) -> None:
    """Links a service to a model.

    Args:
        service_id: The ID of the service to link to the model.
        model: The model to link the service to.
        model_version_id: The ID of the model version to link the service to.

    Raises:
        RuntimeError: If no model is provided and the model context cannot be
            identified.
    """
    client = Client()

    # If no model is provided, try to get it from the context
    if not model and not model_version_id:
        is_issue = False
        try:
            step_context = get_step_context()
            model = step_context.model
        except StepContextError:
            is_issue = True

        if model is None or is_issue:
            raise RuntimeError(
                "`link_service_to_model` called without `model` parameter "
                "and configured model context cannot be identified. Consider "
                "passing the `model` explicitly or configuring it in "
                "@step or @pipeline decorator."
            )

    model_version_id = (
        model_version_id or model._get_or_create_model_version().id
        if model
        else None
    )
    update_service = ServiceUpdate(model_version_id=model_version_id)
    client.zen_store.update_service(
        service_id=service_id, update=update_service
    )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Page model definitions."""

from typing import Generator, Generic, List, TypeVar

from pydantic import BaseModel
from pydantic.types import NonNegativeInt, PositiveInt

from zenml.models.v2.base.filter import BaseFilter

B = TypeVar("B", bound=BaseModel)


class Page(BaseModel, Generic[B]):
    """Return Model for List Models to accommodate pagination."""

    index: PositiveInt
    max_size: PositiveInt
    total_pages: NonNegativeInt
    total: NonNegativeInt
    items: List[B]

    __params_type__ = BaseFilter

    @property
    def size(self) -> int:
        """Return the item count of the page.

        Returns:
            The amount of items in the page.
        """
        return len(self.items)

    def __len__(self) -> int:
        """Return the item count of the page.

        This enables `len(page)`.

        Returns:
            The amount of items in the page.
        """
        return len(self.items)

    def __getitem__(self, index: int) -> B:
        """Return the item at the given index.

        This enables `page[index]`.

        Args:
            index: The index to get the item from.

        Returns:
            The item at the given index.
        """
        return self.items[index]

    def __iter__(self) -> Generator[B, None, None]:  # type: ignore[override]
        """Return an iterator over the items in the page.

        This enables `for item in page` loops, but breaks `dict(page)`.

        Yields:
            An iterator over the items in the page.
        """
        for item in self.items.__iter__():
            yield item

    def __contains__(self, item: B) -> bool:
        """Returns whether the page contains a specific item.

        This enables `item in page` checks.

        Args:
            item: The item to check for.

        Returns:
            Whether the item is in the page.
        """
        return item in self.items



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the MLflow model deployer pipeline step."""

from typing import Optional, cast

from mlflow.tracking import MlflowClient, artifact_utils

from zenml import get_step_context, step
from zenml.artifacts.unmaterialized_artifact import UnmaterializedArtifact
from zenml.client import Client
from zenml.constants import (
    DEFAULT_SERVICE_START_STOP_TIMEOUT,
    MLFLOW_MODEL_FORMAT,
)
from zenml.integrations.mlflow.experiment_trackers.mlflow_experiment_tracker import (
    MLFlowExperimentTracker,
)
from zenml.integrations.mlflow.mlflow_utils import build_runs_uri, is_mlflow_3x
from zenml.integrations.mlflow.model_deployers.mlflow_model_deployer import (
    MLFlowModelDeployer,
)
from zenml.integrations.mlflow.model_registries.mlflow_model_registry import (
    MLFlowModelRegistry,
)
from zenml.integrations.mlflow.services.mlflow_deployment import (
    MLFlowDeploymentConfig,
    MLFlowDeploymentService,
)
from zenml.logger import get_logger
from zenml.model_registries.base_model_registry import (
    ModelRegistryModelMetadata,
    ModelVersionStage,
)

logger = get_logger(__name__)


@step(enable_cache=False)
def mlflow_model_deployer_step(
    model: UnmaterializedArtifact,
    deploy_decision: bool = True,
    experiment_name: Optional[str] = None,
    run_name: Optional[str] = None,
    model_name: str = "model",
    workers: int = 1,
    mlserver: bool = False,
    timeout: int = DEFAULT_SERVICE_START_STOP_TIMEOUT,
) -> Optional[MLFlowDeploymentService]:
    """Model deployer pipeline step for MLflow.

    This step deploys a model logged in the MLflow artifact store to a
    deployment service. The user would typically use this step in a pipeline
    that deploys a model that was already registered in the MLflow model
    registr either manually or by using the `mlflow_model_registry_step`.

    Args:
        deploy_decision: whether to deploy the model or not
        model: the model artifact to deploy
        experiment_name: Name of the MLflow experiment in which the model was
            logged.
        run_name: Name of the MLflow run in which the model was logged.
        model_name: the name of the MLflow model logged in the MLflow artifact
            store for the current pipeline.
        workers: number of workers to use for the prediction service
        mlserver: set to True to use the MLflow MLServer backend (see
            https://github.com/SeldonIO/MLServer). If False, the
            MLflow built-in scoring server will be used.
        timeout: the number of seconds to wait for the service to start/stop.

    Returns:
        MLflow deployment service or None if no service was deployed or found

    Raises:
        ValueError: if the MLflow experiment tracker is not found
    """
    model_deployer = cast(
        MLFlowModelDeployer, MLFlowModelDeployer.get_active_model_deployer()
    )

    experiment_tracker = Client().active_stack.experiment_tracker

    if not isinstance(experiment_tracker, MLFlowExperimentTracker):
        raise ValueError(
            "MLflow model deployer step requires an MLflow experiment "
            "tracker. Please add an MLflow experiment tracker to your "
            "stack."
        )

    # get pipeline name, step name and run id
    step_context = get_step_context()
    pipeline_name = step_context.pipeline.name
    run_name = step_context.pipeline_run.name
    step_name = step_context.step_run.name

    # Configure Mlflow so the client points to the correct store
    experiment_tracker.configure_mlflow()
    client = MlflowClient()
    mlflow_run_id = experiment_tracker.get_run_id(
        experiment_name=experiment_name or pipeline_name,
        run_name=run_name or run_name,
    )

    # Fetch the model URI from the MLflow artifact store
    model_uri = ""
    if mlflow_run_id and client.list_artifacts(mlflow_run_id, model_name):
        # In MLflow 3.x, use runs:/ URI which correctly resolves to model location
        # In MLflow 2.x, use the artifact URI directly
        if is_mlflow_3x():
            model_uri = build_runs_uri(mlflow_run_id, model_name)
        else:
            model_uri = artifact_utils.get_artifact_uri(  # type: ignore[no-untyped-call, unused-ignore]
                run_id=mlflow_run_id, artifact_path=model_name
            )

    predictor_cfg = MLFlowDeploymentConfig(
        model_name=model_name or "",
        model_uri=model_uri,
        workers=workers,
        mlserver=mlserver,
        pipeline_name=pipeline_name,
        pipeline_step_name=step_name,
        timeout=timeout,
    )

    # Fetch existing services with same pipeline name, step name and model name
    existing_services = model_deployer.find_model_server(
        config=predictor_cfg.model_dump(),
    )

    # Check whether to deploy a new service
    if model_uri and deploy_decision:
        new_service = cast(
            MLFlowDeploymentService,
            model_deployer.deploy_model(
                replace=True,
                config=predictor_cfg,
                timeout=timeout,
                service_type=MLFlowDeploymentService.SERVICE_TYPE,
            ),
        )
        logger.info(
            f"MLflow deployment service started and reachable at:\n"
            f"    {new_service.prediction_url}\n"
        )
        logger.info("Stopping existing services...")
        for existing_service in existing_services:
            existing_service.stop(timeout)
        return new_service

    # Check whether an existing service can be reused
    elif existing_services:
        if not model_uri:
            logger.info(
                f"Skipping model deployment because an MLflow model with name "
                f" `{model_name}` was not trained in the current pipeline run."
            )
        elif not deploy_decision:
            logger.info(
                "Skipping model deployment because the deployment decision "
                "was negative."
            )
        logger.info(
            f"Reusing last model server deployed by step `{step_name}` of "
            f"pipeline '{pipeline_name}' for model '{model_name}'..."
        )
        service = cast(MLFlowDeploymentService, existing_services[0])
        if not service.is_running:
            service.start(timeout)
        return service

    # Log a warning if no service was deployed or found
    elif deploy_decision and not model_uri:
        logger.warning(
            f"An MLflow model with name `{model_name}` was not "
            f"logged in the current pipeline run and no running MLflow "
            f"model server was found. Please ensure that your pipeline "
            f"includes a step with an MLflow experiment configured that "
            "trains a model and logs it to MLflow. This could also happen "
            "if the current pipeline run did not log an MLflow model  "
            f"because the training step was cached."
        )
    return None


@step(enable_cache=False)
def mlflow_model_registry_deployer_step(
    registry_model_name: str,
    registry_model_version: Optional[str] = None,
    registry_model_stage: Optional[ModelVersionStage] = None,
    replace_existing: bool = True,
    model_name: str = "model",
    workers: int = 1,
    mlserver: bool = False,
    timeout: int = DEFAULT_SERVICE_START_STOP_TIMEOUT,
) -> MLFlowDeploymentService:
    """Model deployer pipeline step for MLflow.

    Args:
        registry_model_name: the name of the model in the model registry
        registry_model_version: the version of the model in the model registry
        registry_model_stage: the stage of the model in the model registry
        replace_existing: Whether to create a new deployment service or not
        model_name: the name of the MLflow model logged in the MLflow artifact
            store for the current pipeline.
        workers: number of workers to use for the prediction service
        mlserver: set to True to use the MLflow MLServer backend (see
            https://github.com/SeldonIO/MLServer). If False, the
            MLflow built-in scoring server will be used.
        timeout: the number of seconds to wait for the service to start/stop.

    Returns:
        MLflow deployment service

    Raises:
        ValueError: if neither registry_model_version nor registry_model_stage
            is not provided
        ValueError: if No MLflow experiment tracker is found in the current
            active stack
        LookupError: if no model version is found in the MLflow model registry.
    """
    if not registry_model_version and not registry_model_stage:
        raise ValueError(
            "Either registry_model_version or registry_model_stage must "
            "be provided in addition to registry_model_name to the MLflow "
            "model registry deployer step. Since the "
            "mlflow_model_registry_deployer_step is used in conjunction with "
            "the mlflow_model_registry."
        )

    model_deployer = cast(
        MLFlowModelDeployer, MLFlowModelDeployer.get_active_model_deployer()
    )

    # fetch the MLflow model registry
    model_registry = Client().active_stack.model_registry
    if not isinstance(model_registry, MLFlowModelRegistry):
        raise ValueError(
            "The MLflow model registry step can only be used with an "
            "MLflow model registry."
        )

    # fetch the model version
    model_version = None
    if registry_model_version:
        try:
            model_version = model_registry.get_model_version(
                name=registry_model_name,
                version=registry_model_version,
            )
        except KeyError:
            model_version = None
    elif registry_model_stage:
        model_version = model_registry.get_latest_model_version(
            name=registry_model_name,
            stage=registry_model_stage,
        )
    if not model_version:
        raise LookupError(
            f"No Model Version found for model name "
            f"{registry_model_name} and version "
            f"{registry_model_version} or stage "
            f"{registry_model_stage}"
        )
    if model_version.model_format != MLFLOW_MODEL_FORMAT:
        raise ValueError(
            f"Model version {model_version.version} of model "
            f"{model_version.registered_model.name} is not an MLflow model."
            f"Only MLflow models can be deployed with the MLflow deployer "
            f"using this step."
        )
    # fetch existing services with same pipeline name, step name and model name

    existing_services = (
        model_deployer.find_model_server(
            model_name=registry_model_name,
            model_version=model_version.version,
        )
        if replace_existing
        else None
    )
    # create a config for the new model service
    metadata = model_version.metadata or ModelRegistryModelMetadata()
    predictor_cfg = MLFlowDeploymentConfig(
        name=model_name or None,
        model_name=registry_model_name,
        model_version=model_version.version,
        model_uri=model_version.model_source_uri,
        workers=workers,
        mlserver=mlserver,
        pipeline_name=metadata.zenml_pipeline_name or "",
        pipeline_step_name=metadata.zenml_step_name or "",
        timeout=timeout,
    )

    # create a new model deployment and replace an old one if it exists
    new_service = cast(
        MLFlowDeploymentService,
        model_deployer.deploy_model(
            replace=True,
            config=predictor_cfg,
            timeout=timeout,
            service_type=MLFlowDeploymentService.SERVICE_TYPE,
        ),
    )

    logger.info(
        f"MLflow deployment service started and reachable at:\n"
        f"    {new_service.prediction_url}\n"
    )

    if existing_services:
        logger.info("Stopping existing services...")
        for existing_service in existing_services:
            existing_service.stop(timeout)

    return new_service



"""Step run versioning [360fa84718bf].

Revision ID: 360fa84718bf
Revises: 85289fea86ff
Create Date: 2025-07-03 16:11:14.013214

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "360fa84718bf"
down_revision = "85289fea86ff"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.add_column(sa.Column("version", sa.Integer(), nullable=True))
        batch_op.add_column(
            sa.Column("is_retriable", sa.Boolean(), nullable=True)
        )
        batch_op.drop_constraint(
            "unique_step_name_for_pipeline_run", type_="unique"
        )
        batch_op.create_unique_constraint(
            "unique_step_name_for_pipeline_run",
            ["name", "pipeline_run_id", "version"],
        )

    op.execute(
        sa.text("UPDATE step_run SET version = 1 WHERE version IS NULL")
    )
    op.execute(
        sa.text(
            "UPDATE step_run SET is_retriable = FALSE WHERE is_retriable IS NULL"
        )
    )

    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.alter_column(
            "version", existing_type=sa.Integer(), nullable=False
        )
        batch_op.alter_column(
            "is_retriable", existing_type=sa.Boolean(), nullable=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision.

    Raises:
        NotImplementedError: Downgrade not supported.
    """
    # ### commands auto generated by Alembic - please adjust! ###
    raise NotImplementedError("Downgrade not supported")

    # ### end Alembic commands ###



"""Release [0.80.1].

Revision ID: 0.80.1
Revises: 1f1105204aaf
Create Date: 2025-03-27 14:31:15.476115

"""

# revision identifiers, used by Alembic.
revision = "0.80.1"
down_revision = "1f1105204aaf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Carried over version of some functions from the pydantic v1 typing module.

Check out the latest version here:
https://github.com/pydantic/pydantic/blob/v1.10.14/pydantic/typing.py
"""

import sys
import typing
from typing import Any, Optional, Set, Tuple, Type, Union, cast
from typing import get_args as _typing_get_args
from typing import get_origin as _typing_get_origin

from typing_extensions import Annotated, Literal

# Annotated[...] is implemented by returning an instance of one of these
# classes, depending on python/typing_extensions version.
AnnotatedTypeNames = {"AnnotatedMeta", "_AnnotatedAlias"}

# None types
NONE_TYPES: Tuple[Any, Any, Any] = (None, None.__class__, Literal[None])

# Literal types
LITERAL_TYPES: Set[Any] = {Literal}
if hasattr(typing, "Literal"):
    LITERAL_TYPES.add(typing.Literal)

# ----- is_none_type -----


def is_none_type(type_: Any) -> bool:
    """Checks if the provided type is a none type.

    Args:
        type_: type to check.

    Returns:
        boolean indicating whether the type is a none type.
    """
    return type_ in NONE_TYPES


# ----- is_union -----

if sys.version_info < (3, 10):

    def is_union(type_: Optional[Type[Any]]) -> bool:
        """Checks if the provided type is a union type.

        Args:
            type_: type to check.

        Returns:
            boolean indicating whether the type is union type.
        """
        return type_ is Union  # type: ignore[comparison-overlap]


else:

    def is_union(type_: Optional[Type[Any]]) -> bool:
        """Checks if the provided type is a union type.

        Args:
            type_: type to check.

        Returns:
            boolean indicating whether the type is union type.
        """
        import types

        return type_ is Union or type_ is types.UnionType  # type: ignore[comparison-overlap]


# ----- literal -----


def is_literal_type(type_: Type[Any]) -> bool:
    """Checks if the provided type is a literal type.

    Args:
        type_: type to check.

    Returns:
        boolean indicating whether the type is union type.
    """
    return Literal is not None and get_origin(type_) in LITERAL_TYPES


def literal_values(type_: Type[Any]) -> Tuple[Any, ...]:
    """Fetches the literal values defined in a type.

    Args:
        type_: type to check.

    Returns:
        tuple of the literal values.
    """
    return get_args(type_)


def all_literal_values(type_: Type[Any]) -> Tuple[Any, ...]:
    """Fetches the literal values defined in a type in a recursive manner.

    This method is used to retrieve all Literal values as Literal can be
    used recursively (see https://www.python.org/dev/peps/pep-0586)
    e.g. `Literal[Literal[Literal[1, 2, 3], "foo"], 5, None]`

    Args:
        type_: type to check.

    Returns:
        tuple of all the literal values defined in the type.
    """
    if not is_literal_type(type_):
        return (type_,)

    values = literal_values(type_)
    return tuple(x for value in values for x in all_literal_values(value))


# ----- get_origin -----


def get_origin(tp: Type[Any]) -> Optional[Type[Any]]:
    """Fetches the origin of a given type.

    We can't directly use `typing.get_origin` since we need a fallback to
    support custom generic classes like `ConstrainedList`
    It should be useless once https://github.com/cython/cython/issues/3537 is
    solved and https://github.com/pydantic/pydantic/pull/1753 is merged.

    Args:
        tp: type to check

    Returns:
        the origin type of the provided type.
    """
    if type(tp).__name__ in AnnotatedTypeNames:
        return cast(Type[Any], Annotated)  # mypy complains about _SpecialForm
    return _typing_get_origin(tp) or getattr(tp, "__origin__", None)


# ----- get_args -----


def _generic_get_args(tp: Type[Any]) -> Tuple[Any, ...]:
    """Generic get args function.

    In python 3.9, `typing.Dict`, `typing.List`, ...
    do have an empty `__args__` by default (instead of the generic ~T
    for example). In order to still support `Dict` for example and consider
    it as `Dict[Any, Any]`, we retrieve the `_nparams` value that tells us
    how many parameters it needs.

    Args:
        tp: type to check.

    Returns:
        Tuple of all the args.
    """
    if hasattr(tp, "_nparams"):
        return (Any,) * tp._nparams  # type: ignore[no-any-return]
    # Special case for `tuple[()]`, which used to return ((),) with
    # `typing.Tuple in python 3.10- but now returns () for `tuple` and `Tuple`.
    try:
        if tp == Tuple[()] or tp == tuple[()]:  # type: ignore[comparison-overlap]
            return ((),)
    # there is a TypeError when compiled with cython
    except TypeError:  # pragma: no cover
        pass
    return ()


def get_args(tp: Type[Any]) -> Tuple[Any, ...]:
    """Get type arguments with all substitutions performed.

    For unions, basic simplifications used by Union constructor are performed.
    Examples::
        get_args(Dict[str, int]) == (str, int)
        get_args(int) == ()
        get_args(Union[int, Union[T, int], str][int]) == (int, str)
        get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
        get_args(Callable[[], T][int]) == ([], int)

    Args:
        tp: the type to check.

    Returns:
        Tuple of all the args.
    """
    if type(tp).__name__ in AnnotatedTypeNames:
        return tp.__args__ + tp.__metadata__  # type: ignore[no-any-return]
    # the fallback is needed for the same reasons as `get_origin` (see above)
    return (
        _typing_get_args(tp)
        or getattr(tp, "__args__", ())
        or _generic_get_args(tp)
    )


def is_optional(tp: Type[Any]) -> bool:
    """Checks whether a given annotation is typing.Optional.

    Args:
        tp: the type to check.

    Returns:
        boolean indicating if the type is typing.Optional.
    """
    return is_union(get_origin(tp)) and type(None) in get_args(tp)



"""Release [0.63.0].

Revision ID: 0.63.0
Revises: 7d1919bb1ef0
Create Date: 2024-07-29 16:59:59.891939

"""

# revision identifiers, used by Alembic.
revision = "0.63.0"
down_revision = "7d1919bb1ef0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for running pipelines."""

import time
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Union,
)
from uuid import UUID

from pydantic import BaseModel

from zenml.client import Client
from zenml.config.pipeline_run_configuration import PipelineRunConfiguration
from zenml.config.source import Source, SourceType
from zenml.config.step_configurations import StepConfigurationUpdate
from zenml.enums import ExecutionStatus
from zenml.logger import get_logger
from zenml.models import (
    FlavorFilter,
    LogsRequest,
    PipelineRunRequest,
    PipelineRunResponse,
    PipelineRunTriggerInfo,
    PipelineSnapshotBase,
    PipelineSnapshotResponse,
    StackResponse,
)
from zenml.stack import Flavor, Stack
from zenml.utils import (
    code_utils,
    notebook_utils,
    source_utils,
    string_utils,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.base_zen_store import BaseZenStore

if TYPE_CHECKING:
    StepConfigurationUpdateOrDict = Union[
        Dict[str, Any], StepConfigurationUpdate
    ]

logger = get_logger(__name__)


def get_default_run_name(pipeline_name: str) -> str:
    """Gets the default name for a pipeline run.

    Args:
        pipeline_name: Name of the pipeline which will be run.

    Returns:
        Run name.
    """
    return f"{pipeline_name}-{{date}}-{{time}}"


def create_placeholder_run(
    snapshot: "PipelineSnapshotResponse",
    orchestrator_run_id: Optional[str] = None,
    logs: Optional["LogsRequest"] = None,
    trigger_info: Optional[PipelineRunTriggerInfo] = None,
    original_run_id: Optional[UUID] = None,
) -> "PipelineRunResponse":
    """Create a placeholder run for the snapshot.

    Args:
        snapshot: The snapshot for which to create the placeholder run.
        orchestrator_run_id: The orchestrator run ID for the run.
        logs: The logs for the run.
        trigger_info: The trigger information for the run.
        original_run_id: The original run ID for a replayed run.

    Returns:
        The placeholder run.
    """
    start_time = utc_now()
    run_request = PipelineRunRequest(
        name=string_utils.format_name_template(
            name_template=snapshot.run_name_template,
            substitutions=snapshot.pipeline_configuration.finalize_substitutions(
                start_time=start_time,
            ),
        ),
        # We set the start time on the placeholder run already to
        # make it consistent with the {time} placeholder in the
        # run name. This means the placeholder run will usually
        # have longer durations than scheduled runs, as for them
        # the start_time is only set once the first step starts
        # running.
        start_time=start_time,
        orchestrator_run_id=orchestrator_run_id,
        project=snapshot.project_id,
        snapshot=snapshot.id,
        status=ExecutionStatus.INITIALIZING,
        tags=snapshot.pipeline_configuration.tags,
        logs=logs,
        trigger_info=trigger_info,
        original_run_id=original_run_id,
    )
    run, _ = Client().zen_store.get_or_create_run(run_request)
    return run


def wait_for_pipeline_run_to_finish(run_id: UUID) -> "PipelineRunResponse":
    """Waits until a pipeline run is finished.

    Args:
        run_id: ID of the run for which to wait.

    Returns:
        Model of the finished run.
    """
    sleep_interval = 1
    max_sleep_interval = 64

    while True:
        run = Client().get_pipeline_run(run_id)

        if run.status.is_finished:
            return run

        logger.info(
            "Waiting for pipeline run with ID %s to finish (current status: %s)",
            run_id,
            run.status,
        )
        time.sleep(sleep_interval)
        if sleep_interval < max_sleep_interval:
            sleep_interval *= 2


def validate_stack_is_runnable_from_server(
    zen_store: BaseZenStore, stack: StackResponse
) -> None:
    """Validate if a stack model is runnable from the server.

    Args:
        zen_store: ZenStore to use for listing flavors.
        stack: The stack to validate.

    Raises:
        ValueError: If the stack has components of a custom flavor or local
            components.
    """
    for component_list in stack.components.values():
        assert len(component_list) == 1
        component = component_list[0]
        flavors = zen_store.list_flavors(
            FlavorFilter(name=component.flavor_name, type=component.type)
        )
        assert len(flavors) == 1
        flavor_model = flavors[0]

        if flavor_model.is_custom:
            raise ValueError(
                "Unable to run pipeline from the server on a stack that "
                "includes stack components with a custom flavor."
            )

        flavor = Flavor.from_model(flavor_model)
        component_config = flavor.config_class(**component.configuration)

        if component_config.is_local:
            raise ValueError(
                "Unable to run pipeline from the server on a stack that "
                "includes local stack components."
            )


def validate_run_config_is_runnable_from_server(
    run_configuration: "PipelineRunConfiguration",
    is_dynamic: bool,
) -> None:
    """Validates that the run configuration can be used to run from the server.

    Args:
        run_configuration: The run configuration to validate.
        is_dynamic: Whether the snapshot to run is dynamic.

    Raises:
        ValueError: If there are values in the run configuration that are not
            allowed when running a pipeline from the server.
    """
    if run_configuration.parameters and not is_dynamic:
        raise ValueError(
            "Can't set pipeline parameters when running a static pipeline via "
            "the REST API. You can either use dynamic pipelines for which you "
            "can pass pipeline parameters, or refactore your pipeline code to "
            "use step parameters instead of pipeline parameters. For example, "
            "instead of: "
            "```yaml "
            "parameters: "
            "  param1: 1 "
            "  param2: 2 "
            "``` "
            "You'll need to modify your pipeline code to pass parameters "
            "directly to steps: "
            "```yaml "
            "steps: "
            "  step1: "
            "    parameters: "
            "      param1: 1 "
            "      param2: 2 "
            "``` "
        )

    if run_configuration.build:
        raise ValueError("Can't set build when running pipeline via Rest API.")

    if run_configuration.schedule:
        raise ValueError(
            "Can't set schedule when running pipeline via Rest API."
        )

    if run_configuration.settings and run_configuration.settings.get("docker"):
        raise ValueError(
            "Can't set DockerSettings when running pipeline via Rest API."
        )

    if run_configuration.steps:
        for step_update in run_configuration.steps.values():
            if step_update.settings and step_update.settings.get("docker"):
                raise ValueError(
                    "Can't set DockerSettings when running pipeline via "
                    "Rest API."
                )


def upload_notebook_cell_code_if_necessary(
    snapshot: "PipelineSnapshotBase", stack: "Stack"
) -> None:
    """Upload notebook cell code if necessary.

    This function checks if any of the steps of the pipeline that will be
    executed in a different process are defined in a notebook. If that is the
    case, it will extract that notebook cell code into python files and upload
    an archive of all the necessary files to the artifact store.

    Args:
        snapshot: The snapshot.
        stack: The stack on which the snapshot will happen.

    Raises:
        RuntimeError: If the code for one of the steps that will run out of
            process cannot be extracted into a python file.
    """
    should_upload = False
    resolved_notebook_sources = source_utils.get_resolved_notebook_sources()

    for step in snapshot.step_configurations.values():
        source = step.spec.source

        if source.type == SourceType.NOTEBOOK:
            if (
                stack.orchestrator.flavor != "local"
                or step.config.step_operator
            ):
                should_upload = True
                cell_code = resolved_notebook_sources.get(
                    source.import_path, None
                )

                # Code does not run in-process, which means we need to
                # extract the step code into a python file
                if not cell_code:
                    raise RuntimeError(
                        f"Unable to run step {step.config.name}. This step is "
                        "defined in a notebook and you're trying to run it "
                        "in a remote environment, but ZenML was not able to "
                        "detect the step code in the notebook. To fix "
                        "this error, define your step in a python file instead "
                        "of a notebook."
                    )

    if should_upload:
        logger.info("Uploading notebook code...")

        for _, cell_code in resolved_notebook_sources.items():
            notebook_utils.warn_about_notebook_cell_magic_commands(
                cell_code=cell_code
            )
            module_name = notebook_utils.compute_cell_replacement_module_name(
                cell_code=cell_code
            )
            file_name = f"{module_name}.py"

            code_utils.upload_notebook_code(
                artifact_store=stack.artifact_store,
                cell_code=cell_code,
                file_name=file_name,
            )

        all_snapshot_sources = get_all_sources_from_value(snapshot)

        for source in all_snapshot_sources:
            if source.type == SourceType.NOTEBOOK:
                setattr(source, "artifact_store_id", stack.artifact_store.id)

        logger.info("Upload finished.")


def get_all_sources_from_value(value: Any) -> List[Source]:
    """Get all source objects from a value.

    Args:
        value: The value from which to get all the source objects.

    Returns:
        List of source objects for the given value.
    """
    sources = []
    if isinstance(value, Source):
        sources.append(value)
    elif isinstance(value, BaseModel):
        for v in value.__dict__.values():
            sources.extend(get_all_sources_from_value(v))
    elif isinstance(value, Dict):
        for v in value.values():
            sources.extend(get_all_sources_from_value(v))
    elif isinstance(value, (List, Set, tuple)):
        for v in value:
            sources.extend(get_all_sources_from_value(v))

    return sources



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Step execution utilities."""

import time
from typing import (
    TYPE_CHECKING,
    Optional,
)

from zenml.config.step_configurations import Step
from zenml.exceptions import RunStoppedException
from zenml.logger import get_logger
from zenml.models import (
    PipelineSnapshotResponse,
)
from zenml.models.v2.core.step_run import StepRunResponse
from zenml.orchestrators.step_launcher import StepLauncher

if TYPE_CHECKING:
    from zenml.config.step_configurations import Step


logger = get_logger(__name__)


def launch_step(
    snapshot: "PipelineSnapshotResponse",
    step: "Step",
    orchestrator_run_id: str,
    retry: bool = False,
    remaining_retries: Optional[int] = None,
    wait: bool = True,
) -> StepRunResponse:
    """Launch a step.

    Args:
        snapshot: The snapshot.
        step: The step to run.
        orchestrator_run_id: The orchestrator run ID.
        retry: Whether to retry the step if it fails.
        remaining_retries: The number of remaining retries. If not passed, this
            will be read from the step configuration.
        wait: Whether to wait for the step to complete.

    Raises:
        RunStoppedException: If the run was stopped.
        BaseException: If the step failed all retries.

    Returns:
        The step run response.
    """

    def _launch_without_retry() -> StepRunResponse:
        launcher = StepLauncher(
            snapshot=snapshot,
            step=step,
            orchestrator_run_id=orchestrator_run_id,
            wait=wait,
        )
        return launcher.launch()

    if not retry:
        step_run = _launch_without_retry()
    else:
        retries = 0
        retry_config = step.config.retry
        if remaining_retries is None:
            max_retries = retry_config.max_retries if retry_config else 0
        else:
            max_retries = remaining_retries
        delay = retry_config.delay if retry_config else 0
        backoff = retry_config.backoff if retry_config else 1

        while retries <= max_retries:
            try:
                step_run = _launch_without_retry()
            except RunStoppedException:
                # Don't retry if the run was stopped
                raise
            except BaseException:
                retries += 1
                if retries <= max_retries:
                    logger.info(
                        "Sleeping for %d seconds before retrying step `%s`.",
                        delay,
                        step.config.name,
                    )
                    time.sleep(delay)
                    delay *= backoff
                else:
                    if max_retries > 0:
                        logger.error(
                            "Failed to run step `%s` after %d retries.",
                            step.config.name,
                            max_retries,
                        )
                    raise
            else:
                break

    return step_run



# Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base classes for Trigger implementations."""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Literal, Optional, Type

from pydantic import BaseModel, Field, field_validator, model_validator

from zenml.constants import STR_FIELD_MAX_LENGTH
from zenml.enums import TriggerFlavor, TriggerRunConcurrency, TriggerType
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.filter import AnyQuery, BaseFilter
from zenml.models.v2.base.scoped import (
    ProjectScopedFilter,
    ProjectScopedRequest,
    ProjectScopedResponse,
    ProjectScopedResponseBody,
    ProjectScopedResponseMetadata,
    ProjectScopedResponseResources,
)

if TYPE_CHECKING:
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotResponse,
        UserResponse,
    )
    from zenml.models.v2.base.filter import AnySchema


class TriggerBase(BaseModel, ABC):
    """Base class for triggers."""

    name: str = Field(
        max_length=STR_FIELD_MAX_LENGTH,
        description="The name of the trigger.",
    )
    active: bool
    type: TriggerType
    concurrency: TriggerRunConcurrency = Field(
        default=TriggerRunConcurrency.SKIP,
        description="How to handle concurrently running triggers "
        "(pipeline runs generated from the same trigger & snapshot).",
    )


class TriggerRequest(ProjectScopedRequest, TriggerBase, ABC):
    """Base class for trigger requests."""

    flavor: TriggerFlavor

    @abstractmethod
    def get_config(self) -> str:
        """Calculates a serialized JSON object for the type-specific fields of the trigger.

        Returns:
            A JSON string representing the trigger-type configuration.
        """
        pass

    @abstractmethod
    def get_extra_fields(self) -> dict[str, Any]:
        """Calculates extra, type-specific fields needed for the trigger.

        Returns:
            A dictionary of extra fields (e.g. {"next_occurrence": "..."}
        """
        pass

    def get_analytics_metadata(self) -> Dict[str, Any]:
        """Get the analytics metadata for the model.

        Returns:
            Dict of analytics metadata.
        """
        metadata = super().get_analytics_metadata()
        metadata["type"] = self.type.value
        metadata["flavor"] = self.flavor.value
        return metadata


class TriggerUpdate(TriggerBase, BaseUpdate, ABC):
    """Base trigger update class."""

    @abstractmethod
    def get_config(self) -> str:
        """Calculates a serialized JSON object for the type-specific fields of the trigger.

        Returns:
            A JSON string representing the trigger-type configuration.
        """
        pass

    @abstractmethod
    def get_extra_fields(self) -> dict[str, Any]:
        """Calculates extra, type-specific fields needed for the trigger.

        Returns:
            A dictionary of extra fields (e.g. {"next_occurrence": "..."}) to be stored in the database.
            IMPORTANT: These fields must match the attributes defined in the `TriggerSchema` ORM class.
        """
        pass


class TriggerResponseBody(ProjectScopedResponseBody, TriggerBase, ABC):
    """Response body for triggers."""

    is_archived: bool
    flavor: TriggerFlavor

    @abstractmethod
    def get_extra_fields(self) -> list[str]:
        """Specify the extra fields needed for the trigger.

        Returns:
            A list of extra fields (e.g. ["next_occurrence"]).
        """
        pass


class TriggerResponseMetadata(ProjectScopedResponseMetadata):
    """Response metadata for triggers."""

    pass


class TriggerResponseResources(ProjectScopedResponseResources):
    """Class for all resource models associated with the schedule entity."""

    snapshots: Optional[list["PipelineSnapshotResponse"]] = None
    user: Optional["UserResponse"] = None
    latest_run: Optional["PipelineRunResponse"] = None


class UnScopedTriggerFilter(BaseFilter):
    """Base class for filtering triggers."""

    name: str | None = Field(
        default=None,
        description="The name of the trigger.",
    )
    active: bool | None = Field(
        default=None,
        description="Whether the trigger should be active.",
    )
    is_archived: bool = Field(
        default=False,
        description="Whether the trigger should be archived.",
    )
    flavor: TriggerFlavor | str | None = Field(
        default=None,
        description="The trigger flavor.",
        union_mode="left_to_right",
    )
    type: TriggerType | str | None = Field(
        default=None,
        description="The trigger type.",
        union_mode="left_to_right",
    )
    next_occurrence: datetime | str | None = Field(
        default=None,
        description="The next occurrence of the trigger (applicable only for schedules).",
        union_mode="left_to_right",
    )
    concurrency: TriggerRunConcurrency | None = Field(
        default=None, description="The trigger concurrency."
    )


class TriggerFilter(UnScopedTriggerFilter, ProjectScopedFilter):
    """Public class for filtering triggers."""

    FILTER_EXCLUDE_FIELDS: ClassVar[list[str]] = [
        *ProjectScopedFilter.FILTER_EXCLUDE_FIELDS,
        "pipeline_id",
        "snapshot_id",
    ]

    CLI_EXCLUDE_FIELDS: ClassVar[list[str]] = [
        *ProjectScopedFilter.CLI_EXCLUDE_FIELDS,
        "type",
    ]

    pipeline_id: str | None = Field(
        default=None,
        description="Filter triggers by pipeline ID (triggers that are attached to this pipeline's snapshots)",
    )
    snapshot_id: str | None = Field(
        default=None,
        description="Filter triggers by snapshot ID (triggers that are attached to this snapshot)",
    )

    @property
    def filter_by_snapshot(self) -> bool:
        """Implements the `filter_by_snapshot` property.

        Returns:
            True if filtering requires snapshot fields else False.
        """
        return any(f is not None for f in [self.pipeline_id, self.snapshot_id])

    def apply_filter(
        self, query: AnyQuery, table: Type["AnySchema"]
    ) -> AnyQuery:
        """Applies the filter to a query.

        Args:
            query: The query to which to apply the filter.
            table: The query table.

        Returns:
            The query with filter applied.
        """
        from zenml.zen_stores.schemas import (
            PipelineSnapshotSchema,
            TriggerSchema,
            TriggerSnapshotSchema,
        )

        query = ProjectScopedFilter.apply_filter(self, query, table)

        if self.filter_by_snapshot:
            query = query.join(
                TriggerSnapshotSchema,
                TriggerSchema.id == TriggerSnapshotSchema.trigger_id,
            ).join(
                PipelineSnapshotSchema,
                TriggerSnapshotSchema.snapshot_id == PipelineSnapshotSchema.id,
            )

            if self.pipeline_id is not None:
                query = query.where(
                    PipelineSnapshotSchema.pipeline_id == self.pipeline_id
                )

            if self.snapshot_id is not None:
                query = query.where(
                    TriggerSnapshotSchema.snapshot_id == self.snapshot_id
                )

        return query


class ScheduleTrigger(BaseModel):
    """Base class for schedule-specific parameters."""

    cron_expression: str | None = None
    interval: int | None = Field(
        default=None,
        ge=60,
        description="Scheduling option: Execute on intervals of N seconds after start time.",
    )
    end_time: datetime | None = None
    start_time: datetime | None = None
    run_once_start_time: datetime | None = Field(
        default=None,
        description="Scheduling option: Execute once on selected start time.",
    )

    @field_validator(
        "start_time", "end_time", "run_once_start_time", mode="after"
    )
    @classmethod
    def _tz_unaware_utc(cls, value: datetime | None) -> datetime | None:
        """Ensures that all datetime objects are timezone unaware and in UTC time.

        Args:
            value: The datetime.

        Returns:
            The datetime in UTC time without timezone.
        """
        if value and value.tzinfo:
            value = value.astimezone(timezone.utc)
            value = value.replace(tzinfo=None)

        return value

    @model_validator(mode="after")
    def _interval_combined_with_start(self) -> "ScheduleTrigger":
        """Ensures interval is combined with start time.

        Returns:
            The ScheduleTrigger instance.

        Raises:
            ValueError: If interval is set without a start time.
        """
        if self.interval is not None and self.start_time is None:
            raise ValueError("Interval must be set with start time.")

        return self

    @model_validator(mode="after")
    def _unique_scheduling_option(
        self,
    ) -> "ScheduleTrigger":
        """Ensures exactly one frequency configuration option is provided.

        Returns:
            The ScheduleTrigger instance.

        Raises:
            ValueError: If zero or more than one frequency options are provided.
        """
        num_of_set_options = sum(
            1 if option is not None else 0
            for option in [
                self.cron_expression,
                self.interval,
                self.run_once_start_time,
            ]
        )

        if num_of_set_options == 0:
            raise ValueError(
                "You must specify an execution option (cron, interval, one-off) for the schedule."
            )
        if num_of_set_options > 1:
            raise ValueError(
                "You must specify exactly one option (cron, interval, one-off) for the schedule."
            )

        return self

    @model_validator(mode="after")
    def _correct_time_boundaries(self) -> "ScheduleTrigger":
        """Ensure start and end times set valid boundaries.

        Returns:
            The ScheduleTrigger instance.

        Raises:
            ValueError: If start time is greater than end time.
        """
        if not (self.end_time and self.start_time):
            return self

        if self.end_time <= self.start_time:
            raise ValueError("The end time must be after the start time.")

        return self


class ScheduleTriggerRequest(TriggerRequest, ScheduleTrigger):
    """Class representing a ScheduleTrigger request."""

    def get_config(self) -> str:
        """Returns the serialized blob of custom trigger fields.

        Returns:
            Json-dump of ScheduleTrigger config.
        """
        return self.model_dump_json(
            include=set(ScheduleTrigger.model_fields),
        )

    def get_extra_fields(self) -> dict[str, Any]:
        """Calculates and returns extra flat fields.

        Returns:
            A dictionary with extra fields (next occurrence or empty).
        """
        if self.flavor == TriggerFlavor.NATIVE_SCHEDULE:
            from zenml.utils.native_schedules import calculate_first_occurrence

            return {
                "next_occurrence": calculate_first_occurrence(
                    cron_expression=self.cron_expression,
                    start_time=self.start_time,
                    interval=self.interval,
                    run_once_start_time=self.run_once_start_time,
                )
            }
        return {}


class ScheduleTriggerUpdate(TriggerUpdate, ScheduleTrigger):
    """Class representing a ScheduleTrigger update."""

    type: Literal[TriggerType.SCHEDULE] = TriggerType.SCHEDULE
    flavor: Literal[TriggerFlavor.NATIVE_SCHEDULE] = (
        TriggerFlavor.NATIVE_SCHEDULE
    )
    next_occurrence: datetime | None = None

    def get_config(self) -> str:
        """Returns the serialized blob of custom trigger fields.

        Returns:
            Json-dump of ScheduleTrigger config.
        """
        return self.model_dump_json(include=set(ScheduleTrigger.model_fields))

    def get_extra_fields(self) -> dict[str, Any]:
        """Calculates and returns extra flat fields.

        Returns:
            A dictionary with extra fields (next occurrence or empty).
        """
        occurrence_altering_values = [
            self.start_time,
            self.cron_expression,
            self.interval,
        ]

        if not self.flavor == TriggerFlavor.NATIVE_SCHEDULE:
            return {}

        if self.next_occurrence is not None:
            return {
                "next_occurrence": self.next_occurrence,
            }

        if any(field is not None for field in occurrence_altering_values):
            from zenml.utils.native_schedules import calculate_first_occurrence

            return {
                "next_occurrence": calculate_first_occurrence(
                    cron_expression=self.cron_expression,
                    start_time=self.start_time,
                    interval=self.interval,
                    run_once_start_time=self.run_once_start_time,
                )
            }

        return {}


class ScheduleTriggerResponseBody(ScheduleTrigger, TriggerResponseBody):
    """Class representing a ScheduleTrigger response body."""

    next_occurrence: datetime | None = None

    @model_validator(mode="after")
    def _next_occurrence_for_inactive(self) -> "ScheduleTriggerResponseBody":
        if not (self.active and not self.is_archived):
            self.next_occurrence = None

        return self

    def get_extra_fields(self) -> list[str]:
        """Returns the extra fields (e.g. next occurrence).

        Returns:
            The extra fields for the schedule payload.
        """
        return ["next_occurrence"]


class ScheduleTriggerResponse(
    ProjectScopedResponse[
        ScheduleTriggerResponseBody,
        TriggerResponseMetadata,
        TriggerResponseResources,
    ]
):
    """Class representing a ScheduleTrigger response."""

    @property
    def is_archived(self) -> bool:
        """Implements the 'is_archived' property.

        Returns:
            True if the trigger is archived.
        """
        return self.get_body().is_archived

    @property
    def active(self) -> bool:
        """Implements the 'is_active' property.

        Returns:
            True if the trigger is active.
        """
        return self.get_body().active

    @property
    def name(self) -> str:
        """Implements the 'name' property.

        Returns:
            The trigger name.
        """
        return self.get_body().name

    @property
    def type(self) -> TriggerType:
        """Implements the 'type' property.

        Returns:
            The trigger type.
        """
        return self.get_body().type

    @property
    def flavor(self) -> TriggerFlavor:
        """Implements the 'flavor' property.

        Returns:
            The trigger flavor.
        """
        return self.get_body().flavor

    @property
    def next_occurrence(self) -> datetime | None:
        """Implements the 'next_occurrence' property.

        Returns:
            The schedule's next occurrence.
        """
        return self.get_body().next_occurrence

    @property
    def cron_expression(self) -> str | None:
        """Implements the 'cron_expression' property.

        Returns:
            The schedule's cron expression.
        """
        return self.get_body().cron_expression

    @property
    def start_time(self) -> datetime | None:
        """Implements the 'start_time' property.

        Returns:
            The schedule's start time.
        """
        return self.get_body().start_time

    @property
    def interval(self) -> int | None:
        """Implements the 'interval' property.

        Returns:
            The schedule's interval.
        """
        return self.get_body().interval

    @property
    def end_time(self) -> datetime | None:
        """Implements the 'end_time' property.

        Returns:
            The schedule's end time.
        """
        return self.get_body().end_time

    @property
    def run_once_start_time(self) -> datetime | None:
        """Implements the 'run_once_start_time' property.

        Returns:
            The schedule's run once start time.
        """
        return self.get_body().run_once_start_time

    @property
    def snapshots(self) -> list["PipelineSnapshotResponse"] | None:
        """Implements the 'snapshots' property.

        Returns:
            A list of snapshots the triggers is attached to.
        """
        return self.get_resources().snapshots

    @property
    def latest_run(self) -> Optional["PipelineRunResponse"]:
        """Implements the 'latest_run' property.

        Returns:
            The latest run of the trigger.
        """
        return self.get_resources().latest_run

    @property
    def concurrency(self) -> TriggerRunConcurrency:
        """Implements the 'concurrency' property.

        Returns:
            The concurrency of the trigger.
        """
        return self.get_body().concurrency


TRIGGER_UPDATE_TYPE_UNION = ScheduleTriggerUpdate
TRIGGER_CREATE_TYPE_UNION = ScheduleTriggerRequest
TRIGGER_RETURN_TYPE_UNION = ScheduleTriggerResponse



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""AzureML definitions."""

from typing import Optional

from pydantic import Field, model_validator

from zenml.config.base_settings import BaseSettings
from zenml.logger import get_logger
from zenml.utils.enum_utils import StrEnum

logger = get_logger(__name__)


class AzureMLComputeTypes(StrEnum):
    """Enum for different types of compute on AzureML."""

    SERVERLESS = "serverless"
    COMPUTE_INSTANCE = "compute-instance"
    COMPUTE_CLUSTER = "compute-cluster"


class AzureMLComputeSettings(BaseSettings):
    """Settings for the AzureML compute.

    These settings adjust the compute resources that will be used by the
    pipeline execution.

    There are three possible use cases for this implementation:

        1. Serverless compute (default behavior):
            - The `mode` is set to `serverless` (default behavior).
            - All the other parameters become irrelevant and will throw a
            warning if set.

        2. Compute instance:
            - The `mode` is set to `compute-instance`.
            - In this case, users have to provide a `compute-name`.
                - If a compute instance exists with this name, this instance
                will be used and all the other parameters become irrelevant
                and will throw a warning if set.
                - If a compute instance does not already exist, ZenML will
                create it. You can use the parameters `compute_size` and
                `idle_type_before_shutdown_minutes` for this operation.

        3. Compute cluster:
            - The `mode` is set to `compute-cluster`.
            - In this case, users have to provide a `compute-name`.
                - If a compute cluster exists with this name, this instance
                will be used and all the other parameters become irrelevant
                and will throw a warning if set.
                - If a compute cluster does not already exist, ZenML will
                create it. You can set the additional parameters for this
                operation.
    """

    # Mode for compute
    mode: AzureMLComputeTypes = AzureMLComputeTypes.SERVERLESS

    # Common Configuration for Compute Instances and Clusters
    compute_name: Optional[str] = None
    size: Optional[str] = None

    # Additional configuration for a Compute Instance
    idle_time_before_shutdown_minutes: Optional[int] = None

    # Additional configuration for a Compute Cluster
    idle_time_before_scaledown_down: Optional[int] = None
    location: Optional[str] = None
    min_instances: Optional[int] = None
    max_instances: Optional[int] = None
    tier: Optional[str] = None

    # Container resource configuration
    shm_size: Optional[str] = Field(
        default=None,
        description="Size of the shared memory block for the container. Must be "
        "specified in format supported by Docker (e.g., '2g', '200g', '1024m'). "
        "Useful for PyTorch dataloaders with multiple workers, large model caching, "
        "and RAM-based operations. Defaults to AzureML's default (typically 64MB)",
    )

    @model_validator(mode="after")
    def azureml_settings_validator(self) -> "AzureMLComputeSettings":
        """Checks whether the right configuration is set based on mode.

        Returns:
            the instance itself.

        Raises:
            AssertionError: if a name is not provided when working with
                instances and clusters.
        """
        viable_configuration_fields = {
            AzureMLComputeTypes.SERVERLESS: {"mode", "shm_size"},
            AzureMLComputeTypes.COMPUTE_INSTANCE: {
                "mode",
                "compute_name",
                "size",
                "idle_time_before_shutdown_minutes",
                "shm_size",
            },
            AzureMLComputeTypes.COMPUTE_CLUSTER: {
                "mode",
                "compute_name",
                "size",
                "idle_time_before_scaledown_down",
                "location",
                "min_instances",
                "max_instances",
                "tier",
                "shm_size",
            },
        }
        viable_fields = viable_configuration_fields[self.mode]

        for field in self.model_fields_set:
            if (
                field not in viable_fields
                and field in AzureMLComputeSettings.model_fields
            ):
                logger.warning(
                    f"In the {self.__class__.__name__} settings, the mode of "
                    f"operation is set to {self.mode}. In this mode, you can "
                    f"not configure the parameter '{field}'. This "
                    "configuration will be ignored."
                )

        if (
            self.mode == AzureMLComputeTypes.COMPUTE_INSTANCE
            or self.mode == AzureMLComputeTypes.COMPUTE_CLUSTER
        ):
            assert self.compute_name is not None, (
                "When you are working with compute instances and clusters, "
                "please define a name for the compute target."
            )

        return self



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes template engine."""

import json
from typing import Any, Dict, List

import yaml
from jinja2 import (
    Environment,
    StrictUndefined,
    TemplateError,
    Undefined,
    select_autoescape,
)

from zenml.io import fileio
from zenml.logger import get_logger
from zenml.utils import io_utils

logger = get_logger(__name__)


def _to_yaml_filter(obj: Any) -> str:
    """Convert Python object to YAML string.

    Args:
        obj: Python object to convert.

    Returns:
        YAML string representation.
    """
    return yaml.safe_dump(
        obj, default_flow_style=False, allow_unicode=True
    ).rstrip()


def _to_json_filter(obj: Any) -> str:
    """Convert Python object to JSON string.

    Args:
        obj: Python object to convert.

    Returns:
        JSON string representation.
    """
    return json.dumps(obj)


class KubernetesTemplateEngine:
    """Engine for generating Kubernetes resources from Jinja2 templates.

    Attributes:
        env: Jinja2 environment for rendering templates.
    """

    def __init__(
        self,
        strict_undefined: bool = True,
    ):
        """Initialize the template engine.

        Args:
            strict_undefined: If True, raise an error for undefined template
                variables. If False, undefined variables are silently ignored.
        """
        # autoescape must be disabled for YAML manifests rendered from plain
        # strings because select_autoescape defaults to escaping string-based
        # templates as HTML. That would turn `"foo"` into `&#34;foo&#34;`, making
        # the rendered YAML invalid.
        self.env = Environment(
            undefined=StrictUndefined if strict_undefined else Undefined,
            autoescape=select_autoescape(
                enabled_extensions=("html", "htm", "xml"),
                default=False,
                default_for_string=False,
            ),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )

        self.env.filters["to_yaml"] = _to_yaml_filter
        self.env.filters["tojson"] = _to_json_filter

    # ========================================================================
    # Template and Resource Rendering
    # ========================================================================

    def render_template(
        self,
        file: str,
        context: Dict[str, Any],
    ) -> List[Dict[str, Any]]:
        """Render a YAML file with optional Jinja2 template context.

        Args:
            file: Path to the YAML file (local or remote).
            context: Optional Jinja2 template context for variable substitution.

        Returns:
            List of resource dictionaries.

        Raises:
            FileNotFoundError: If the template file doesn't exist.
            ValueError: If template rendering or YAML parsing fails.
        """
        if io_utils.is_remote(file):
            file_path = io_utils.sanitize_remote_path(file)
        else:
            file_path = io_utils.resolve_relative_path(file)

        if not fileio.exists(file_path):
            raise FileNotFoundError(f"Template file not found: {file}")

        file_contents = io_utils.read_file_contents_as_string(file_path)

        try:
            template = self.env.from_string(file_contents)
            yaml_content = template.render(**context)
        except TemplateError as e:
            raise ValueError(
                f"Failed to render Jinja2 template '{file}': {e}"
            ) from e

        return self.load_yaml_documents(yaml_content)

    @staticmethod
    def load_yaml_documents(yaml_content: str) -> List[Dict[str, Any]]:
        """Load one or more YAML documents from a string.

        Supports two common Kubernetes YAML patterns:
        1. Multiple documents separated by `---`:
           ```yaml
           ---
           apiVersion: v1
           kind: ConfigMap
           ---
           apiVersion: apps/v1
           kind: Deployment
           ```
        2. Top-level list of resources:
           ```yaml
           - apiVersion: v1
             kind: ConfigMap
           - apiVersion: apps/v1
             kind: Deployment
           ```

        Args:
            yaml_content: YAML string potentially containing multiple documents.

        Returns:
            List of resource dictionaries (skips None documents).

        Raises:
            ValueError: If YAML is invalid or contains non-dict documents.
        """
        try:
            documents = list(yaml.safe_load_all(yaml_content))
        except yaml.YAMLError as e:
            raise ValueError(f"Invalid YAML: {e}") from e

        resources: List[Dict[str, Any]] = []
        for index, document in enumerate(documents):
            if document is None:
                continue
            if isinstance(document, list):
                for item_index, item in enumerate(document):
                    if item is None:
                        continue
                    if not isinstance(item, dict):
                        raise ValueError(
                            f"YAML document {index + 1}, item {item_index + 1} "
                            f"must be a dictionary, got {type(item).__name__}"
                        )
                    resources.append(item)
            elif isinstance(document, dict):
                resources.append(document)
            else:
                raise ValueError(
                    f"YAML document {index + 1} must be a dictionary or list, "
                    f"got {type(document).__name__}"
                )

        return resources



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of a VertexAI step operator."""

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, cast

from google.cloud import aiplatform

from zenml import __version__
from zenml.config.build_configuration import BuildConfiguration
from zenml.enums import ExecutionStatus, StackComponentType
from zenml.integrations.gcp.constants import (
    VERTEX_ENDPOINT_SUFFIX,
)
from zenml.integrations.gcp.flavors.vertex_step_operator_flavor import (
    VertexStepOperatorConfig,
    VertexStepOperatorSettings,
)
from zenml.integrations.gcp.google_credentials_mixin import (
    GoogleCredentialsMixin,
)
from zenml.integrations.gcp.utils import (
    build_job_request,
    convert_job_state,
)
from zenml.logger import get_logger
from zenml.orchestrators.publish_utils import publish_step_run_metadata
from zenml.stack import Stack, StackValidator
from zenml.step_operators import BaseStepOperator

if TYPE_CHECKING:
    from zenml.config.base_settings import BaseSettings
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import PipelineSnapshotBase, StepRunResponse

logger = get_logger(__name__)

VERTEX_DOCKER_IMAGE_KEY = "vertex_step_operator"
STEP_JOB_NAME_METADATA_KEY = "job_name"


class VertexStepOperator(BaseStepOperator, GoogleCredentialsMixin):
    """Step operator to run a step on Vertex AI.

    This class defines code that can set up a Vertex AI environment and run the
    ZenML entrypoint command in it.
    """

    _job_service_client: Optional[aiplatform.gapic.JobServiceClient] = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the step operator and validates the accelerator type.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

    @property
    def config(self) -> VertexStepOperatorConfig:
        """Returns the `VertexStepOperatorConfig` config.

        Returns:
            The configuration.
        """
        return cast(VertexStepOperatorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Vertex step operator.

        Returns:
            The settings class.
        """
        return VertexStepOperatorSettings

    @property
    def validator(self) -> Optional[StackValidator]:
        """Validates the stack.

        Returns:
            A validator that checks that the stack contains a remote container
            registry and a remote artifact store.
        """

        def _validate_remote_components(stack: "Stack") -> Tuple[bool, str]:
            if stack.artifact_store.config.is_local:
                return False, (
                    "The Vertex step operator runs code remotely and "
                    "needs to write files into the artifact store, but the "
                    f"artifact store `{stack.artifact_store.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote artifact store when using the Vertex "
                    "step operator."
                )

            container_registry = stack.container_registry
            assert container_registry is not None

            if container_registry.config.is_local:
                return False, (
                    "The Vertex step operator runs code remotely and "
                    "needs to push/pull Docker images, but the "
                    f"container registry `{container_registry.name}` of the "
                    "active stack is local. Please ensure that your stack "
                    "contains a remote container registry when using the "
                    "Vertex step operator."
                )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_remote_components,
        )

    def get_docker_builds(
        self, snapshot: "PipelineSnapshotBase"
    ) -> List["BuildConfiguration"]:
        """Gets the Docker builds required for the component.

        Args:
            snapshot: The pipeline snapshot for which to get the builds.

        Returns:
            The required Docker builds.
        """
        builds = []
        for step_name, step in snapshot.step_configurations.items():
            if step.config.uses_step_operator(self.name):
                build = BuildConfiguration(
                    key=VERTEX_DOCKER_IMAGE_KEY,
                    settings=step.config.docker_settings,
                    step_name=step_name,
                )
                builds.append(build)

        return builds

    def get_job_service_client(self) -> aiplatform.gapic.JobServiceClient:
        """Get the job service client.

        Returns:
            The job service client.
        """
        if self.connector_has_expired():
            self._job_service_client = None

        if self._job_service_client is None:
            credentials, _ = self._get_authentication()
            client_options = {
                "api_endpoint": self.config.region + VERTEX_ENDPOINT_SUFFIX
            }
            self._job_service_client = aiplatform.gapic.JobServiceClient(
                credentials=credentials, client_options=client_options
            )
        return self._job_service_client

    def submit(
        self,
        info: "StepRunInfo",
        entrypoint_command: List[str],
        environment: Dict[str, str],
    ) -> None:
        """Submits a step run to VertexAI.

        Args:
            info: Information about the step run.
            entrypoint_command: Command that executes the step.
            environment: Environment variables to set in the step operator
                environment.
        """
        resource_settings = info.config.resource_settings
        if resource_settings.cpu_count or resource_settings.memory:
            logger.warning(
                "Specifying cpus or memory is not supported for "
                "the Vertex step operator. If you want to run this step "
                "operator on specific resources, you can do so by configuring "
                "a different machine_type type like this: "
                "`zenml step-operator update %s "
                "--machine_type=<MACHINE_TYPE>`",
                self.name,
            )
        settings = cast(VertexStepOperatorSettings, self.get_settings(info))
        image = info.get_image(key=VERTEX_DOCKER_IMAGE_KEY)

        labels = {"source": f"zenml-{__version__.replace('.', '_')}"}
        job_request = build_job_request(
            display_name=f"{info.run_name}-{info.pipeline_step_name}",
            image=image,
            entrypoint_command=entrypoint_command,
            custom_job_settings=settings,
            resource_settings=info.config.resource_settings,
            environment=environment,
            labels=labels,
            encryption_spec_key_name=self.config.encryption_spec_key_name,
            service_account=self.config.service_account,
            network=self.config.network,
        )
        logger.debug("Vertex AI Job=%s", job_request)

        client = self.get_job_service_client()
        parent = (
            f"projects/{self.gcp_project_id}/locations/{self.config.region}"
        )
        logger.info(
            "Submitting custom job='%s', path='%s' to Vertex AI Training.",
            job_request["display_name"],
            parent,
        )
        info.force_write_logs()
        response = client.create_custom_job(
            parent=parent, custom_job=job_request
        )
        logger.debug("Vertex AI response:", response)
        publish_step_run_metadata(
            info.step_run_id,
            {self.id: {STEP_JOB_NAME_METADATA_KEY: response.name}},
        )
        info.step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY] = response.name

    def get_status(self, step_run: "StepRunResponse") -> ExecutionStatus:
        """Gets the status of a submitted step.

        Args:
            step_run: The step run.

        Returns:
            The step status.
        """
        client = self.get_job_service_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        job = client.get_custom_job(name=job_name)
        return convert_job_state(job.state)

    def cancel(self, step_run: "StepRunResponse") -> None:
        """Cancels a submitted step.

        Args:
            step_run: The step run.
        """
        client = self.get_job_service_client()
        job_name = step_run.run_metadata[STEP_JOB_NAME_METADATA_KEY]
        client.cancel_custom_job(name=job_name)



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""ZenML server deployment exceptions."""

from zenml.exceptions import ZenMLBaseException


class ServerDeploymentError(ZenMLBaseException):
    """Base exception class for all ZenML server deployment related errors."""


class ServerProviderNotFoundError(ServerDeploymentError):
    """Raised when using a ZenML server provider that doesn't exist."""


class ServerDeploymentExistsError(ServerDeploymentError):
    """Raised when trying to deploy a new ZenML server with the same name."""


class ServerDeploymentNotFoundError(ServerDeploymentError):
    """Raised when trying to fetch a ZenML server deployment that doesn't exist."""


class ServerDeploymentConfigurationError(ServerDeploymentError):
    """Raised when there is a ZenML server deployment configuration error ."""



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Log exporter that writes logs to Datadog."""

from typing import TYPE_CHECKING, Any, Dict, Sequence

from zenml.log_stores.otel.otel_log_exporter import OTLPLogExporter
from zenml.logger import get_logger

DEFAULT_TIMEOUT = 10

if TYPE_CHECKING:
    from opentelemetry.sdk._logs import LogData

logger = get_logger(__name__)


class DatadogLogExporter(OTLPLogExporter):
    """Datadog log exporter.

    This exporter writes OpenTelemetry logs to Datadog's HTTP intake API
    with slightly modified log records to adapt to Datadog's format.
    """

    @classmethod
    def _encode_log(cls, log_data: "LogData") -> Dict[str, Any]:
        """Encode a log data object to a dictionary.

        Args:
            log_data: The log data object to encode.

        Returns:
            A dictionary representing the log data.
        """
        body = log_data.log_record.body
        attributes = (
            dict(log_data.log_record.attributes)
            if log_data.log_record.attributes
            else {}
        )

        resource = log_data.log_record.resource
        instrumentation = log_data.instrumentation_scope

        if resource.attributes:
            attributes.update(dict(resource.attributes))

        if instrumentation.attributes:
            attributes.update(dict(instrumentation.attributes))

        log_record = dict(
            **attributes,
            service=attributes.get("service.name"),
            version=attributes.get("service.version"),
            timestamp=int(log_data.log_record.timestamp / 1e6)
            if log_data.log_record.timestamp
            else None,
            message=str(body),
            status=log_data.log_record.severity_text.lower()
            if log_data.log_record.severity_text
            else "info",
            source="zenml",
        )

        return {k: v for k, v in log_record.items() if v is not None}

    def _encode_logs(self, logs: Sequence["LogData"]) -> Any:
        """Encode a sequence of log data objects to a list of dictionaries.

        Args:
            logs: The sequence of log data objects to encode.

        Returns:
            The log data.
        """
        return [self._encode_log(log) for log in logs]
